# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/relics_rpc.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nbox.relics.proto import relics_pb2 as proto_dot_relics__pb2
from nbox.relics.proto import common_pb2 as proto_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16proto/relics_rpc.proto\x1a\x12proto/relics.proto\x1a\x12proto/common.proto\"\x8a\x01\n\x12\x43reateRelicRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0e\n\x06region\x18\x03 \x01(\t\x12\x1a\n\x04\x61uth\x18\x04 \x01(\x0b\x32\x0c.BackendInfo\x12$\n\x0b\x62ucket_meta\x18\x05 \x01(\x0b\x32\x0f.BucketMetadata\"`\n\x11ListRelicsRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x12\n\nrelic_name\x18\x02 \x01(\t\x12\x0f\n\x07page_no\x18\x03 \x01(\x05\x12\x10\n\x08relic_id\x18\x04 \x01(\t\"B\n\x12ListRelicsResponse\x12\x16\n\x06relics\x18\x01 \x03(\x0b\x32\x06.Relic\x12\x14\n\x0ctotal_relics\x18\x02 \x01(\x05\"\x87\x01\n\x15ListRelicFilesRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x10\n\x08relic_id\x18\x02 \x01(\t\x12\x12\n\nrelic_name\x18\x03 \x01(\t\x12\x0e\n\x06prefix\x18\x04 \x01(\t\x12\x11\n\tfile_name\x18\x06 \x01(\t\x12\x0f\n\x07page_no\x18\x07 \x01(\x05\"H\n\x16ListRelicFilesResponse\x12\x19\n\x05\x66iles\x18\x01 \x03(\x0b\x32\n.RelicFile\x12\x13\n\x0btotal_files\x18\x02 \x01(\x05\"\xa7\x01\n\x12\x41\x63tivityLogRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x10\n\x08relic_id\x18\x02 \x01(\t\x12\x12\n\nrelic_name\x18\x03 \x01(\t\x12\x10\n\x08username\x18\x04 \x01(\t\x12\x19\n\x11\x66rom_timestamp_ns\x18\x05 \x01(\x03\x12\x17\n\x0fto_timestamp_ns\x18\x06 \x01(\x03\x12\x0f\n\x07page_no\x18\x07 \x01(\x05\"\x8e\x01\n\x0b\x41\x63tivityLog\x12\x14\n\x0ctimestamp_ns\x18\x01 \x01(\x03\x12\x10\n\x08relic_id\x18\x02 \x01(\t\x12\x1b\n\toperation\x18\x03 \x01(\x0e\x32\x08.Actions\x12\x13\n\x0bobject_size\x18\x04 \x01(\x02\x12\x10\n\x08username\x18\x05 \x01(\t\x12\x13\n\x0bobject_name\x18\x06 \x01(\t\"1\n\x13\x41\x63tivityLogResponse\x12\x1a\n\x04logs\x18\x01 \x03(\x0b\x32\x0c.ActivityLog\"E\n\x11ListBucketRequest\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x1a\n\x04\x61uth\x18\x02 \x01(\x0b\x32\x0c.BackendInfo\"%\n\x12ListBucketResponse\x12\x0f\n\x07\x62uckets\x18\x01 \x03(\t*7\n\x07\x41\x63tions\x12\x08\n\x04NONE\x10\x00\x12\n\n\x06\x43REATE\x10\x01\x12\n\n\x06UPDATE\x10\x02\x12\n\n\x06\x44\x45LETE\x10\x03\x32\xbb\x04\n\nRelicStore\x12-\n\x0c\x63reate_relic\x12\x13.CreateRelicRequest\x1a\x06.Relic\"\x00\x12\x38\n\x0blist_relics\x12\x12.ListRelicsRequest\x1a\x13.ListRelicsResponse\"\x00\x12+\n\x11update_relic_meta\x12\x06.Relic\x1a\x0c.Acknowledge\"\x00\x12&\n\x0c\x64\x65lete_relic\x12\x06.Relic\x1a\x0c.Acknowledge\"\x00\x12%\n\x11get_relic_details\x12\x06.Relic\x1a\x06.Relic\"\x00\x12\'\n\x0b\x63reate_file\x12\n.RelicFile\x1a\n.RelicFile\"\x00\x12\x45\n\x10list_relic_files\x12\x16.ListRelicFilesRequest\x1a\x17.ListRelicFilesResponse\"\x00\x12\x31\n\x12\x64\x65lete_multi_files\x12\x0b.RelicFiles\x1a\x0c.Acknowledge\"\x00\x12)\n\rdownload_file\x12\n.RelicFile\x1a\n.RelicFile\"\x00\x12?\n\x10get_activity_log\x12\x13.ActivityLogRequest\x1a\x14.ActivityLogResponse\"\x00\x12\x39\n\x0clist_buckets\x12\x12.ListBucketRequest\x1a\x13.ListBucketResponse\"\x00\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'proto.relics_rpc_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ACTIONS._serialized_start=1061
  _ACTIONS._serialized_end=1116
  _CREATERELICREQUEST._serialized_start=67
  _CREATERELICREQUEST._serialized_end=205
  _LISTRELICSREQUEST._serialized_start=207
  _LISTRELICSREQUEST._serialized_end=303
  _LISTRELICSRESPONSE._serialized_start=305
  _LISTRELICSRESPONSE._serialized_end=371
  _LISTRELICFILESREQUEST._serialized_start=374
  _LISTRELICFILESREQUEST._serialized_end=509
  _LISTRELICFILESRESPONSE._serialized_start=511
  _LISTRELICFILESRESPONSE._serialized_end=583
  _ACTIVITYLOGREQUEST._serialized_start=586
  _ACTIVITYLOGREQUEST._serialized_end=753
  _ACTIVITYLOG._serialized_start=756
  _ACTIVITYLOG._serialized_end=898
  _ACTIVITYLOGRESPONSE._serialized_start=900
  _ACTIVITYLOGRESPONSE._serialized_end=949
  _LISTBUCKETREQUEST._serialized_start=951
  _LISTBUCKETREQUEST._serialized_end=1020
  _LISTBUCKETRESPONSE._serialized_start=1022
  _LISTBUCKETRESPONSE._serialized_end=1059
  _RELICSTORE._serialized_start=1119
  _RELICSTORE._serialized_end=1690
# @@protoc_insertion_point(module_scope)


import os

from nbox import utils as U
from nbox import messages as mpb
from nbox.utils import logger, lo
from nbox.auth import secret, auth_info_pb
from nbox.version import __version__
from nbox.init import nbox_grpc_stub, nbox_serving_service_stub

from nbox.nbxlib.astea import Astea, IndexTypes as IT
from nbox.hyperloop.common.common_pb2 import Resource
from nbox.hyperloop.jobs.job_pb2 import (
  Job as JobProto,
)
from nbox.hyperloop.jobs.dag_pb2 import DAG as DAGProto
from nbox.hyperloop.jobs.nbox_ws_pb2 import JobRequest
from nbox.hyperloop.deploy.serve_pb2 import (
  ServingRequest,
  Serving,
)

from nbox.jd_core.jobs import Job
from nbox.jd_core.serving import Serve


def upload_job_folder(
  method: str,
  init_folder: str,
  id: str = "",
  project_id: str = "",

  # job / deploy rpc things
  trigger: bool = False,
  deploy: bool = True,
  pin: bool = False,
  feature_gates: dict = {},

  # all the things for resources
  resource_cpu: str = "",
  resource_memory: str = "",
  resource_disk_size: str = "",
  resource_gpu: str = "",
  resource_gpu_count: str = "",
  resource_timeout: int = 0,
  resource_max_retries: int = 0,

  # deployment specific
  model_name: str = "",

  # X-type
  serving_type: str = "nbox",
  exe: str = "",

  # there's no more need to pass the workspace_id anymore
  workspace_id: str = "",


  # some extra things for functionality
  _ret: bool = False,

  # finally everything else is assumed to be passed to the initialisation script
  **init_kwargs
):
  """Upload the code for a job or serving to the NBX.

  ### Engineer's Note

  This function is supposed to be exposed via CLI, you can of course make a programtic call to this as well. This is a reason
  why this function can keep on taking many arguments. However with greatm number of arguments comes great responsibility, ie.
  lots of if/else conditions. Broadly speaking this should manage all the complexity and pass along simple reduced intstructions
  to the underlying methods. Currently the arguments that are required in Jinja templates are packed as `exe_jinja_kwargs` and
  we call the `deploy_job` and `deploy_serving`.

  Args:
    method (str): The method to use, either "job" or "serving"
    init_folder (str): folder with all the relevant files or ``file_path:fn_name`` pair so you can use it as the entrypoint.
    name (str, optional): Name of the job. Defaults to "".
    id (str, optional): ID of the job. Defaults to "".
    project_id (str, optional): Project ID, if None uses the one from config. Defaults to "".
    trigger (bool, optional): If uploading a "job" trigger the job after uploading. Defaults to False.
    resource_cpu (str, optional): CPU resource. Defaults to "100m".
    resource_memory (str, optional): Memory resource. Defaults to "128Mi".
    resource_disk_size (str, optional): Disk size resource. Defaults to "3Gi".
    resource_gpu (str, optional): GPU resource. Defaults to "none".
    resource_gpu_count (str, optional): GPU count resource. Defaults to "0".
    resource_timeout (int, optional): Timeout resource. Defaults to 120_000.
    resource_max_retries (int, optional): Max retries resource. Defaults to 2.
    cron (str, optional): Cron string for scheduling. Defaults to "".
    workspace_id (str, optional): Workspace ID, if None uses the one from config. Defaults to "".
    init_kwargs (dict): kwargs to pass to the `init` function / class, if possible
  """
  if exe:
    logger.warning(f"--exe is a new feature, please report any bugs to the nbox team")

  from nbox.network import deploy_job, deploy_serving
  import nbox.nbxlib.operator_spec as ospec
  from nbox.nbxlib.serving import SupportedServingTypes as SST
  from nbox.projects import Project
  
  OT = ospec.OperatorType

  if method not in OT._valid_deployment_types():
    raise ValueError(f"Invalid method: {method}, should be either {OT._valid_deployment_types()}")
  if trigger and method != OT.JOB:
    raise ValueError(f"Trigger can only be used with method='{OT.JOB}'")
  if pin and method != OT.SERVING:
    raise ValueError(f"Deploy and Pin can only be used with method='{OT.SERVING}'")
  if model_name and method != OT.SERVING:
    raise ValueError(f"model_name can only be used with '{OT.SERVING}'")
  
  # get the correct ID based on the project_id
  if (not project_id and not id) or (project_id and id):
    raise ValueError("Either --project-id or --id must be present")
  if project_id:
    p = Project(project_id)
    if method == OT.JOB:
      id = p.get_job_id()
    else:
      id = p.get_deployment_id()
    logger.info(f"Using project_id: {project_id}, found id: {id}")

  if ":" not in init_folder:
    # this means we are uploading a traditonal folder that contains a `nbx_user.py` file
    # in this case the module is loaded on the local machine and so user will need to have
    # everything installed locally. This was a legacy method before 0.10.0
    logger.error(
      'Old method of having a manual nbx_user.py file is now deprecated\n'
      f'  Fix: nbx {method} upload file_path:fn_cls_name --id "id"'
    )
    raise ValueError("Old style upload is not supported anymore")

  # In order to upload we can either chose to upload the entire folder, but can we implement a style where
  # only that specific function is uploaded? This is useful for raw distributed compute style.
  commands = init_folder.split(":")
  if len(commands) == 2:
    fn_file, fn_name = commands
    mode = "folder"
  elif len(commands) == 3:
    mode, fn_file, fn_name = commands
    if mode not in ["file", "folder"]:
      raise ValueError(f"Invalid mode: '{mode}' in upload command, should be either 'file' or 'folder'")
  else:
    raise ValueError(f"Invalid init_folder: {init_folder}")
  if mode != "folder":
    raise NotImplementedError(f"Only folder mode is supported, got: {mode}")
  if not os.path.exists(fn_file+".py"):
    raise ValueError(f"File {fn_file}.py does not exist")
  init_folder, file_name = os.path.split(fn_file)
  init_folder = init_folder or "."
  fn_name = fn_name.strip()
  if not os.path.exists(init_folder):
    raise ValueError(f"Folder {init_folder} does not exist")
  logger.info(f"Uploading code from folder: {init_folder}:{file_name}:{fn_name}")
  _curdir = os.getcwd()
  os.chdir(init_folder)

  workspace_id = workspace_id or secret.workspace_id

  # Now that we have the init_folder and function name, we can throw relevant errors
  perform_tea = True
  if method == OT.SERVING:
    if serving_type not in SST.all():
      raise ValueError(f"Invalid serving_type: {serving_type}, should be one of {SST.all()}")
    if serving_type == SST.FASTAPI or serving_type == SST.FASTAPI_V2:
      logger.warning(f"You have selected serving_type='{serving_type}', this assumes the object: {fn_name} is a FastAPI app")
      init_code = fn_name
      perform_tea = False
      load_operator = False

  if perform_tea:
    # build an Astea and analyse it for getting the computation that is going to be run
    load_operator = True
    tea = Astea(fn_file+".py")
    items = tea.find(fn_name, [IT.CLASS, IT.FUNCTION])
    if len(items) > 1:
      raise ModuleNotFoundError(f"Multiple {fn_name} found in {fn_file}.py")
    elif len(items) == 0:
      logger.error(f"Could not find function or class type: '{fn_name}'")
      raise ModuleNotFoundError(f"Could not find function or class type: '{fn_name}'")
    fn = items[0]
    if fn.type == IT.FUNCTION:
      # does not require initialisation
      if len(init_kwargs):
        logger.error(
          f"Cannot pass kwargs to a function: '{fn_name}'\n"
          f"  Fix: you cannot pass kwargs {set(init_kwargs.keys())} to a function"
        )
        raise ValueError("Function does not require initialisation")
      init_code = fn_name
    elif fn.type == IT.CLASS:
      # requires initialisation, in this case we will store the relevant to things in a Relic
      init_comm = ",".join([f"{k}={v}" for k, v in init_kwargs.items()])
      init_code = f"{fn_name}({init_comm})"
      logger.info(f"Starting with init code:\n  {init_code}")

  # load up the things that are to be passed to the exe.py file
  exe_jinja_kwargs = {
    "file_name": file_name,
    "fn_name": fn_name,
    "init_code": init_code,
    "load_operator": load_operator,
    "serving_type": serving_type,
  }

  # create a requirements.txt file if it doesn't exist with the latest nbox version
  if not os.path.exists(U.join(".", "requirements.txt")):
    with open(U.join(".", "requirements.txt"), "w") as f:
      f.write(f'nbox[serving]=={__version__}\n')

  # create a .nboxignore file if it doesn't exist, this will tell all the files not to be uploaded to the job pod
  _igp = set(ospec.FN_IGNORE)
  if not os.path.exists(U.join(".", ".nboxignore")):
    with open(U.join(".", ".nboxignore"), "w") as f:
      f.write("\n".join(_igp))
  else:
    with open(U.join(".", ".nboxignore"), "r") as f:
      _igp = _igp.union(set(f.read().splitlines()))
    _igp = sorted(list(_igp)) # sort it so it doesn't keep creating diffs in the git
    with open(U.join(".", ".nboxignore"), "w") as f:
      f.write("\n".join(_igp))

  # creation of resources, we first need to check if any resource arguments are passed, if they are
  def __common_resource(db: Resource) -> Resource:
    # get a common resource based on what the user has said, what the db has and defaults if nothing is given
    logger.debug(lo("db resources", **mpb.message_to_dict(db)))
    resource = Resource(
      cpu = str(resource_cpu) or db.cpu or ospec.DEFAULT_RESOURCE.cpu,
      memory = str(resource_memory) or db.memory or ospec.DEFAULT_RESOURCE.memory,
      disk_size = str(resource_disk_size) or db.disk_size or ospec.DEFAULT_RESOURCE.disk_size,
      gpu = str(resource_gpu) or db.gpu or ospec.DEFAULT_RESOURCE.gpu,
      gpu_count = str(resource_gpu_count) or db.gpu_count or ospec.DEFAULT_RESOURCE.gpu_count,
      timeout = int(resource_timeout) or db.timeout or ospec.DEFAULT_RESOURCE.timeout,
      max_retries = int(resource_max_retries) or db.max_retries or ospec.DEFAULT_RESOURCE.max_retries,
    )
    return resource

  # common to both, kept out here because these two will eventually merge
  nbx_auth_info = auth_info_pb()
  if method == ospec.OperatorType.JOB:
    # since user has not passed any arguments, we will need to check if the job already exists
    job_proto: JobProto = nbox_grpc_stub.GetJob(
      JobRequest(
        auth_info = nbx_auth_info,
        job = JobProto(id = id)
      )
    )
    resource = __common_resource(job_proto.resource)
    logger.debug(lo("resources", **mpb.message_to_dict(resource)))
    out: Job = deploy_job(
      init_folder = init_folder,
      job_id = job_proto.id,
      job_name = job_proto.name,
      feature_gates = feature_gates,
      dag = DAGProto(),
      workspace_id = workspace_id,
      schedule = None,
      resource = resource,
      exe_jinja_kwargs = exe_jinja_kwargs,
      exe = exe,
    )
    if trigger:
      logger.info(f"Triggering job: {job_proto.name} ({job_proto.id})")
      out = out.trigger()

  elif method == ospec.OperatorType.SERVING:
    model_name = model_name or U.get_random_name().replace("-", "_")
    logger.info(f"Model name: {model_name}")

    # serving_id, serving_name = _get_deployment_data(name = name, id = id, workspace_id = workspace_id)
    serving_proto: Serving = nbox_serving_service_stub.GetServing(
      ServingRequest(
        auth_info = nbx_auth_info,
        serving = Serving(id=id),
      )
    )
    resource = __common_resource(serving_proto.resource)
    logger.debug(lo("resources", **mpb.message_to_dict(resource)))
    out: Serve = deploy_serving(
      init_folder = init_folder,
      serving_id = serving_proto.id,
      model_name = model_name,
      serving_name = serving_proto.name,
      workspace_id = workspace_id,
      resource = resource,
      wait_for_deployment = False,
      model_metadata = {
        "serving_type": serving_type
      },
      exe_jinja_kwargs = exe_jinja_kwargs,
      exe = exe,
    )
    if deploy:
      out.deploy(feature_gates = feature_gates, resource = resource)
    if trigger:
      out.pin()
  else:
    raise ValueError(f"Unknown method: {method}")

  os.chdir(_curdir)

  if _ret:
    return out


"""
Utility objects and functions.

This has a couple of cool things:

- `get_logger`: a master logger for nbox, this can be modified to log through anything
- `isthere`: a decorator that checks if a package is installed, if not it raises an error\
  It is more complicated than it needs to be because it is seedling for a way to package\
  functions and code together so that it can be used in a more dynamic way.
- `get_files_in_folder`: a function that returns all files in a folder with a certain extension
- `fetch`: a function that fetches a url and caches it in `tempdir` for faster loading
- `get_random_name`: a function that returns a random name, if `True` is passed returns\
  an `uuid4()` for truly random names :)
- `hash_`: a function that returns a hash of any python object, string is accurate, others\
  might be anything, but atleast it returns something.
- `folder/join`: to be used in pair, `join(folder(__file__), "server_temp.jinja")` means\
  relative path "../other.py". Scattered throughout the codebase, the super generic name will\
  bite me.
- `to_pickle/from_pickle`: to be used in pair, `to_pickle(obj, "path")` and `from_pickle("path")`\
  to save and load python objects to disk.
- `DBase`: सस्ता-protobuf (cheap protobuf), can be nested and `get_dict` will get for all\
  children
- `PoolBranch`: Or how to use multiprocessing, but blocked so you don't have to give a shit\
  about it.
"""

############################################################
# This file is d0 meaning that this has no dependencies!
# Do not import anything from rest of nbox here!
# only exception is `nbox.nbxlib.logger`
############################################################

# this file has bunch of functions that are used everywhere

import os
import sys
import json
import uuid
import hashlib
import requests
import tempfile
import traceback
import randomname
import cloudpickle
from uuid import uuid4
from typing import List, Any
from contextlib import contextmanager
from base64 import b64encode, b64decode
from datetime import datetime, timezone
from functools import partial, lru_cache
from importlib.util import spec_from_file_location, module_from_spec
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, as_completed
from google.protobuf.timestamp_pb2 import Timestamp as Timestamp_pb

from nbox.nbxlib.logger import get_logger, lo

class env:
  """
  Single namespace for all environment variables.
  
  #. `NBOX_LOG_LEVEL`: Logging level for `nbox`, set `NBOX_LOG_LEVEL=info|debug|warning"
  #. `NBOX_ACCESS_TOKEN`: User token for NimbleBox.ai from [secrets](app.nimblebox.ai/secrets)
  #. `NBOX_NO_AUTH`: If set `secrets = None`, this will break any API request and is good only for local testing
  #. `NBOX_NO_LOAD_GRPC`: If set, will not load grpc stub
  #. `NBOX_NO_LOAD_WS`: If set, will not load webserver subway
  #. `NBOX_NO_CHECK_VERSION`: If set, will not check for version
  #. `NBOX_SSH_NO_HOST_CHECKING`: If set, `ssh` will not check for host key
  #. `NBOX_HOME_DIR`: By default `~/.nbx` folder, avoid changing this, user generally does not need to set this
  #. `NBOX_JSON_LOG`: Whether to print json-logs, user generally does not need to set this
  #. `NBOX_JOB_FOLDER`: Folder path for the job, user generally does not need to set this
  """
  # things user can chose to set if they want
  NBOX_LOG_LEVEL = lambda x: os.getenv("NBOX_LOG_LEVEL", x)
  NBOX_ACCESS_TOKEN = lambda x: os.getenv("NBOX_ACCESS_TOKEN", x)
  NBOX_SSH_NO_HOST_CHECKING = lambda x: os.getenv("NBOX_SSH_NO_HOST_CHECKING", x)

  # things that are good mostly for testing and development of nbox itself
  NBOX_NO_AUTH = lambda x: os.getenv("NBOX_NO_AUTH", x)
  NBOX_NO_LOAD_GRPC = lambda: os.getenv("NBOX_NO_LOAD_GRPC", False)
  NBOX_NO_LOAD_WS = lambda: os.getenv("NBOX_NO_LOAD_WS", False)
  NBOX_NO_CHECK_VERSION = lambda: os.getenv("NBOX_NO_CHECK_VERSION", False)
  
  # that that should be avoided to change, but provided here for max. control
  NBOX_HOME_DIR = lambda : os.environ.get("NBOX_HOME_DIR", os.path.join(os.path.expanduser("~"), ".nbx"))
  NBOX_JSON_LOG = lambda x: os.getenv("NBOX_JSON_LOG", x)
  NBOX_JOB_FOLDER = lambda x: os.getenv("NBOX_JOB_FOLDER", x)

  def set(key, value):
    os.environ[key] = value

  def get(key, default=None):
    return os.environ.get(key, default)

# logger /
logger = get_logger(env) # package wide logger
# logger = NBXUnifiedLogger(env) # package wide logger

def log_traceback():
  # for l in traceback.format_exc().splitlines():
  #   logger.error(l)
  logger.error(traceback.format_exc())

@contextmanager
def deprecation_warning(msg, remove, replace_by: str = None, help: str = None):
  from nbox.version import __version__
  msg = "Deprecation Warning" \
    f"\n  current: {__version__}" \
    f"\n  removed: {remove}" \
    f"\n      msg: {msg}"
  if replace_by:
    msg += f"\n  replace: {replace_by}"
  if help:
    msg += f"\n  help: {help}"
  logger.warning(msg)

class FileLogger:
  """Flush logs to a file, useful when we don't want to mess with current logging"""
  def __init__(self, filepath):
    self.filepath = filepath
    self.f = open(filepath, "a+")

    self.debug = partial(self.log, level="debug",)
    self.info = partial(self.log, level="info",)
    self.warning = partial(self.log, level="warning",)
    self.error = partial(self.log, level="error",)
    self.critical = partial(self.log, level="critical",)

  def log(self, message, level):
    self.f.write(f"[{datetime.now(timezone.utc).isoformat()}] {level}: {message}\n")
    self.f.flush()

# / logger

# lazy_loading/

def load_module_from_path(fn_name, file_path):
  spec = spec_from_file_location(fn_name, file_path)
  foo = module_from_spec(spec)
  mod_name = get_random_name()
  sys.modules[mod_name] = foo
  spec.loader.exec_module(foo)
  fn = getattr(foo, fn_name)
  return fn

def isthere(*packages, soft = True):
  """Checks all the packages

  Args:
      soft (bool, optional): If `False` raises `ImportError`. Defaults to `True`.
  """
  def wrapper(fn):
    _fn_ = fn
    def _fn(*args, **kwargs):
      # since we are lazy evaluating this thing, we are checking when the function
      # is actually called. This allows checks not to happen during __init__.
      for package in packages:
        # split the package name to get version number as well, not all will have package
        # version so continue those that have a version
        package = package.split("==")
        if len(package) == 1:
          package_name, package_version = package[0], None
        else:
          package_name, package_version = package
        if package_name in sys.modules:
          # trying to install these using pip will cause issues, so avoid that
          continue

        try:
          module = __import__(package_name)
          if hasattr(module, "__version__") and package_version:
            if module.__version__ != package_version:
              raise ImportError(f"{package_name} version mismatch")
        except ImportError:
          if not soft:
            raise ImportError(f"{package} is not installed, but is required by {fn}")
          # raise a warning, let the modulenotfound exception bubble up
          logger.warning(
            f"{package} is not installed, but is required by {fn.__module__}, some functionality may not work"
          )
      return _fn_(*args, **kwargs)
    return _fn
  return wrapper


# /lazy_loading

# path/

def get_files_in_folder(
  folder,
  ext = ["*"],
  abs_path: bool = True,
  followlinks: bool = False,
) -> List[str]:
  """Get files with `ext` in `folder`"""
  # this method is faster than glob
  import os
  all_paths = []
  _all = "*" in ext # wildcard means everything so speed up

  folder_abs = os.path.abspath(folder) if abs_path else folder
  for root,_,files in os.walk(folder_abs, followlinks=followlinks):
    if _all:
      all_paths.extend([join(root, f) for f in files])
      continue

    for f in files:
      for e in ext:
        if f.endswith(e):
          all_paths.append(os.path.join(root,f))
  return all_paths

def fetch(url, force = False):
  """Fetch and cache a url for faster loading, `force` re-downloads"""
  fp = join(tempfile.gettempdir(), hash_(url))
  if os.path.isfile(fp) and os.stat(fp).st_size > 0 and not force:
    with open(fp, "rb") as f:
      dat = f.read()
  else:
    dat = requests.get(url).content
    with open(fp + ".tmp", "wb") as f:
      f.write(dat)
    os.rename(fp + ".tmp", fp)
  return dat

def folder(x) -> str:
  """get the folder of this file path"""
  return os.path.split(os.path.abspath(x))[0]

def join(x, *args) -> str:
  """convienience function for os.path.join"""
  return os.path.join(x, *args)

def to_pickle(obj, path):
  """Save an object to a pickle file
  
  Args:
    obj: object to save
    path: path to save to
  """
  with open(path, "wb") as f:
    cloudpickle.dump(obj, f)

def from_pickle(path):
  """Load an object from a pickle file

  Args:
    path: path to load from
  """
  with open(path, "rb") as f:
    return cloudpickle.load(f)

def py_to_bs64(x: Any):
  return b64encode(cloudpickle.dumps(x)).decode("utf-8")

def py_from_bs64(x: bytes):
  return cloudpickle.loads(b64decode(x.encode("utf-8")))

def to_json(x: dict, fp: str = "", indent = 2):
  if fp:
    with open(fp, "w") as f:
      f.write(json.dumps(x, indent = indent))
  else:
    return json.dumps(x, indent = indent)

def from_json(x: str):
  if os.path.isfile(x):
    with open(x, "r") as f:
      return json.load(f)
  else:
    return json.loads(x)

def get_assets_folder():
  return join(folder(__file__), "assets")

# /path

# checks/

def is_valid_uuid(x):
  try:
    uuid.UUID(x)
    return True
  except ValueError:
    return False

# /checks

# misc/

def get_random_name(uuid = False):
  """Get a random name, if `uuid` is `True`, return a uuid4"""
  if uuid:
    return str(uuid4())
  return randomname.generate()

def hash_(item, fn="md5"):
  """Hash sting of any item"""
  return getattr(hashlib, fn)(str(item).encode("utf-8")).hexdigest()


@lru_cache()
def get_mime_type(fp: str, defualt = "application/octet-stream"):
  with open(join(get_assets_folder(), "ex2mime.json"), "r") as f:
    mimes = json.load(f)
  return mimes.get(os.path.splitext(fp)[1].lower(), defualt)


# /misc

# datastore/

class DBase:
  def __init__(self, **kwargs):
    for k, v in kwargs.items():
      setattr(self, k, v)

  def get(self, k, v = None):
    return getattr(self, k, v)
  
  def get_dict(self):
    data = {}
    for k in self.__slots__:
      _obj = getattr(self, k, None)
      if _obj == None:
        continue
      if isinstance(_obj, DBase):
        data[k] = _obj.get_dict()
      elif _obj != None and isinstance(_obj, (list, tuple)) and len(_obj) and isinstance(_obj[0], DBase):
        data[k] = [_obj.get_dict() for _obj in _obj]
      else:
        data[k] = _obj
    return data

  def __repr__(self):
    return str(self.get_dict())

  def json(self, fp: str = "") -> str:
    if fp:
      with open(fp, "w") as f:
        f.write(json.dumps(self.get_dict()))
    else:    
      return json.dumps(self.get_dict())

# /datastore


################################################################################
# Parallel
# ========
# There already are many multiprocessing libraries for thread, core, pod, cluster
# but thee classes below are inspired by https://en.wikipedia.org/wiki/Collective_operation
#
# And that is why they are blocking classes, ie. it won't stop till all the tasks
# are completed. For reference please open the link above which has diagrams, there
# nodes can just be threads/cores/... Here is description for each of them:
#
# - Pool: apply same functions on different inputs
# - Branch: apply different functions on different inputs
################################################################################

# pool/

def threaded_map(fn, inputs, wait: bool = True, max_threads = 20, _name: str = None) -> None:
  """
  inputs is a list of tuples, each tuple is the input for single invocation of fn. order is preserved.
  """
  _name = _name or get_random_name(True)
  results = [None for _ in range(len(inputs))]
  with ThreadPoolExecutor(max_workers = max_threads, thread_name_prefix = _name) as exe:
    _fn = lambda i, x: [i, fn(*x)]
    futures = {exe.submit(_fn, i, x): i for i, x in enumerate(inputs)}
    if not wait:
      return futures
    for future in as_completed(futures):
      try:
        i, res = future.result()
        results[i] = res
      except Exception as e:
        raise e
  return results

# /pool

def hard_exit_program(code = 0):
  # why use os._exit over sys.exit:
  # https://stackoverflow.com/questions/9591350/what-is-difference-between-sys-exit0-and-os-exit0
  # https://stackoverflow.com/questions/19747371/python-exit-commands-why-so-many-and-when-should-each-be-used
  # tl;dr: os._exit kills without cleanup and so it's okay on the Pod
  os._exit(code)


class SimplerTimes:
  tz = timezone.utc

  def get_now_datetime() -> datetime:
    return datetime.now(SimplerTimes.tz)

  def get_now_float() -> float:
    return SimplerTimes.get_now_datetime().timestamp()

  def get_now_i64() -> int:
    return int(SimplerTimes.get_now_datetime().timestamp())

  def get_now_str() -> str:
    return SimplerTimes.get_now_datetime().strftime("%Y-%m-%d %H:%M:%S.%f")

  def get_now_pb():
    ts = Timestamp_pb()
    ts.GetCurrentTime()
    return ts
  
  def get_now_ns() -> int:
    return SimplerTimes.get_now_pb().ToNanoseconds()

  def i64_to_datetime(i64) -> datetime:
    return datetime.fromtimestamp(i64, SimplerTimes.tz)



import sys
from datetime import datetime
import tabulate
from functools import lru_cache
from typing import Tuple, List, Dict, Any

from google.protobuf.field_mask_pb2 import FieldMask

from nbox import messages as mpb
from nbox.utils import logger, lo
from nbox.auth import secret, auth_info_pb
from nbox.init import nbox_ws_v1, nbox_grpc_stub

from nbox.hyperloop.jobs.nbox_ws_pb2 import (
  JobRequest,
  ListJobsRequest,
  ListJobsResponse,
  UpdateJobRequest
)
from nbox.hyperloop.jobs.job_pb2 import Job as JobProto

from nbox.jd_core.schedule import Schedule

################################################################################
"""
# NimbleBox.ai Jobs

This is the actual job object that users can manipulate. It is a shallow class
around the NBX-Jobs gRPC API.
"""
################################################################################

@lru_cache()
def _get_job_data(name: str = "", id: str = "", remove_archived: bool = True, *, workspace_id: str = "") -> Tuple[str, str]:
  """
  Returns the job id and name
  
  Args:
    name (str, optional): Job name. Defaults to "".
    id (str, optional): Job ID. Defaults to "".
    remove_archived (bool, optional): If True, will remove archived jobs. Defaults to True.

  Returns:
    Tuple[str, str]: Job ID, Job Name
  """
  if (not name and not id) or (name and id):
    logger.info(f"Please either pass job_id '{id}' or name '{name}'")
    return None, None
  # get stub
  workspace_id = workspace_id or secret.workspace_id
  if workspace_id == None:
    workspace_id = "personal"

  job: JobProto = mpb.rpc(
    nbox_grpc_stub.GetJob,
    JobRequest(
      auth_info = auth_info_pb(),
      job = JobProto(id=id, name=name)
    ),
    "Could not find job with ID: {}".format(id),
    raise_on_error = True
  )
  if job.status == JobProto.Status.ARCHIVED and remove_archived:
    logger.info(f"Job {job.id} is archived. Will try to create a new Job.")
    return None, name
  job_name = job.name
  job_id = job.id
  logger.info(f"Found job with ID '{job_id}' and name '{job_name}'")
  return job_id, job_name

def get_job_list(sort: str = "created_at", *, workspace_id: str = ""):
  """Get list of jobs
  
  Args:
    sort (str, optional): Sort key. Defaults to "name".
  """
  workspace_id = workspace_id or secret.workspace_id

  def _get_time(t):
    return datetime.fromtimestamp(int(float(t))).strftime("%Y-%m-%d %H:%M:%S")

  out: ListJobsResponse = mpb.rpc(
    nbox_grpc_stub.ListJobs,
    ListJobsRequest(auth_info = auth_info_pb()),
    "Could not get job list",
  )

  if len(out.jobs) == 0:
    logger.info("No jobs found")
    sys.exit(0)

  headers = ['created_at', 'id', 'name', 'schedule', 'status']
  try:
    sorted_jobs = sorted(out.jobs, key = lambda x: getattr(x, sort))
  except Exception as e:
    logger.error(f"Cannot sort on key: {sort}")
    sorted_jobs = out.jobs
  data = []
  for j in sorted_jobs:
    _row = []
    for x in headers:
      if x == "status":
        _row.append(JobProto.Status.keys()[getattr(j, x)])
        continue
      elif x == "created_at":
        _row.append(_get_time(j.created_at.seconds))
      elif x == "schedule":
        _row.append(j.schedule.cron)
      else:
        _row.append(getattr(j, x))
    data.append(_row)
  for l in tabulate.tabulate(data, headers).splitlines():
    logger.info(l)


class Job:

  def __init__(self, job_name: str = "", job_id: str = ""):
    """Python wrapper for NBX-Jobs gRPC API, when both arguments are not passed,
    an unintialiased object is created.

    Args:
      job_name (str, optional): Job name. Defaults to "".
      job_id (str, optional): Job ID. Defaults to "".
    """
    if job_name == "" and job_id == "":
      return

    self.id, self.name = _get_job_data(job_name, job_id)
    self.workspace_id = secret.workspace_id
    self.auth_info = auth_info_pb()
    self.job_proto = JobProto(id = self.id)

    self.run_stub = None
    self.runs = []

    # sometimes a Job can be a placeholder and not have any data against it, thus it won't make
    # sense to try to get the data. This causes RPC error
    if self.id is not None:
      self.refresh()
      self.get_runs() # load the runs as well

  @property
  def exists(self):
    """Check if this job exists in the workspace"""
    return self.id is not None

  @classmethod
  def new(cls, job_name: str, description: str = ""):
    """Create a new job

    Args:
      job_name (str): Job name
      description (str, optional): Job description. Defaults to "".

    Returns:
      Job: Job object
    """
    job = nbox_ws_v1.job("post", job_name = job_name, job_description = description)
    job_id = job["job_id"]
    # job_proto: JobProto = nbox_grpc_stub.CreateJob(
    #   JobRequest(
    #     auth_info = auth_info_pb(),
    #     job = JobProto(
    #       name = job_name,
    #       code = Code(
    #         size = 1,
    #         type = Code.Type.NOT_SET,
    #       ),
    #       resource = DEFAULT_RESOURCE,
    #       schedule = JobProto.Schedule(
    #         cron = "0 0 * * *",
    #         end = Timestamp(seconds = 0),
    #       ),
    #       status = JobProto.Status.NOT_SET,
    #       paused = False,
    #     ),
    #   )
    # )
    logger.info(f"Created a new job with name (ID): {job_name} ({job_id})")
    return cls(job_id = job_id)

  def change_schedule(self, new_schedule: Schedule):
    """Change schedule this job
    
    Args:
      new_schedule (Schedule): New schedule
    """
    logger.debug(f"Updating job '{self.job_proto.id}'")
    self.job_proto.schedule.MergeFrom(new_schedule.get_message())
    mpb.rpc(
      nbox_grpc_stub.UpdateJob,
      UpdateJobRequest(auth_info=self.auth_info, job=self.job_proto, update_mask=FieldMask(paths=["schedule"])),
      "Could not update job schedule",
      raise_on_error = True
    )
    logger.debug(f"Updated job '{self.job_proto.id}'")
    self.refresh()

  def __repr__(self) -> str:
    x = f"nbox.Job('{self.job_proto.id}', '{self.job_proto.auth_info.workspace_id}'): {self.status}"
    if self.job_proto.schedule.ByteSize != None:
      x += f" {self.job_proto.schedule}"
    else:
      x += " (no schedule)"
    return x

  def logs(self, f = sys.stdout):
    """Stream logs of the job, `f` can be anything has a `.write/.flush` methods"""
    logger.debug(f"Streaming logs of job '{self.job_proto.id}'")
    for job_log in mpb.streaming_rpc(
      nbox_grpc_stub.GetJobLogs,
      JobRequest(auth_info=self.auth_info ,job = self.job_proto),
      f"Could not get logs of job {self.job_proto.id}, is your job complete?",
      True
    ):
      for log in job_log.log:
        f.write(log)
        f.flush()

  def delete(self):
    """Delete this job"""
    logger.info(f"Deleting job '{self.job_proto.id}'")
    mpb.rpc(nbox_grpc_stub.DeleteJob, JobRequest(auth_info=self.auth_info, job = self.job_proto,), "Could not delete job")
    logger.info(f"Deleted job '{self.job_proto.id}'")
    self.refresh()

  def refresh(self):
    """Refresh Job data"""
    logger.info(f"Updating job '{self.job_proto.id}'")
    if self.id == None:
      self.id, self.name = _get_job_data(id = self.id, workspace_id = self.workspace_id)
    if self.id == None:
      return

    self.job_proto: JobProto = mpb.rpc(
      nbox_grpc_stub.GetJob,
      JobRequest(auth_info=self.auth_info, job = self.job_proto),
      f"Could not get job {self.job_proto.id}"
    )
    self.auth_info.CopyFrom(auth_info_pb())
    logger.debug(f"Updated job '{self.job_proto.id}'")

    self.status = self.job_proto.Status.keys()[self.job_proto.status]

  def trigger(self, tag: str = "", feature_gates: dict = {}):
    """Manually triger this job.
    
    Args:
      tag (str, optional): Tag to be set in the run metadata, read in more detail before trying to use this. Defaults to "".
    """
    logger.debug(f"Triggering job '{self.job_proto.id}'")
    if tag:
      self.job_proto.feature_gates.update(feature_gates)
      self.job_proto.feature_gates.update({"SetRunMetadata": tag})
    mpb.rpc(nbox_grpc_stub.TriggerJob, JobRequest(auth_info=self.auth_info, job = self.job_proto), f"Could not trigger job '{self.job_proto.id}'")
    logger.info(f"Triggered job '{self.job_proto.id}'")
    self.refresh()

  def pause(self):
    """Pause the execution of this job.

    **WARNING: This will "cancel" all the scheduled runs, if present**
    """
    logger.info(f"Pausing job '{self.job_proto.id}'")
    job: JobProto = self.job_proto
    job.status = JobProto.Status.PAUSED
    mpb.rpc(nbox_grpc_stub.UpdateJob, UpdateJobRequest(auth_info=self.auth_info, job=job, update_mask=FieldMask(paths=["status", "paused"])), f"Could not pause job {self.job_proto.id}", True)
    logger.debug(f"Paused job '{self.job_proto.id}'")
    self.refresh()

  def resume(self):
    """Resume the Job with the current schedule, if provided else simlpy sets status as ACTIVE"""
    logger.info(f"Resuming job '{self.job_proto.id}'")
    job: JobProto = self.job_proto
    job.status = JobProto.Status.SCHEDULED
    mpb.rpc(nbox_grpc_stub.UpdateJob, UpdateJobRequest(auth_info=self.auth_info, job=job, update_mask=FieldMask(paths=["status", "paused"])), f"Could not resume job {self.job_proto.id}", True)
    logger.debug(f"Resumed job '{self.job_proto.id}'")
    self.refresh()

  def _get_runs(self, page = -1, limit = 10) -> List[Dict]:
    self.run_stub = nbox_ws_v1.job.u(self.id).runs
    runs = self.run_stub(limit = limit, page = page)["runs_list"]
    return runs

  def get_runs(self, page = -1, sort = "s_no", limit = 10) -> List[Dict]:
    """
    Get runs for this job

    Args:
      page (int, optional): Page number. Defaults to -1.
      sort (str, optional): Sort by. Defaults to "s_no".
      limit (int, optional): Number of runs to return. Defaults to 10.

    Returns:
      List[Dict]: List of runs
    """
    runs = self._get_runs(page, limit)
    sorted_runs = sorted(runs, key = lambda x: x[sort])
    return sorted_runs

  def display_runs(self, sort: str = "created_at", page: int = -1, limit = 10):
    """Display runs for this job

    Args:
      sort (str, optional): Sort by. Defaults to "created_at".
      page (int, optional): Page number. Defaults to -1.
      limit (int, optional): Number of runs to return. Defaults to 10.
    """
    headers = ["s_no", "created_at", "run_id", "status"]

    def _display_runs(runs):
      data = []
      for run in runs:
        data.append((run["s_no"], run["created_at"], run["run_id"], run["status"]))
      for l in tabulate.tabulate(data, headers).splitlines():
        logger.info(l)

    runs = self.get_runs(sort = sort, page = page, limit = limit) # time should be reverse
    _display_runs(runs)
    if page == -1:
      page = 1
    y = input(f">> Print {limit} more runs? (y/n): ")
    done = y != "y"
    while not done:
      _display_runs(self.get_runs(page = page + 1, sort = sort, limit = limit))
      page += 1
      y = input(f">> Print {limit} more runs? (y/n): ")
      done = y != "y"

  def last_n_runs(self, n: int = 10) -> List[Dict]:
    """Get last n runs for this job

    Args:
      n (int, optional): Number of runs to return. Defaults to 10.

    Returns:
      List[Dict]: List of runs
    """
    all_items = []
    _page = 1
    out = self._get_runs(_page, n)
    all_items.extend(out)

    while len(all_items) < n:
      _page += 1
      out = self._get_runs(_page, n)
      if not len(out):
        break
      all_items.extend(out)

    if n == 1:
      return all_items[0]
    return all_items



import sys
import grpc
import datetime
import tabulate
from functools import lru_cache
from typing import Tuple,  Dict, Any

from google.protobuf.field_mask_pb2 import FieldMask

from nbox import messages as mpb
from nbox.utils import logger, lo
from nbox.auth import secret, auth_info_pb
from nbox.init import nbox_ws_v1, nbox_model_service_stub, nbox_serving_service_stub

from nbox.hyperloop.deploy.serve_pb2 import (
  ServingListResponse,
  ServingRequest,
  Serving,
  ServingListRequest,
  UpdateServingRequest,
  ModelRequest,
  Model as ModelProto,
  UpdateModelRequest,
)
from nbox.hyperloop.common.common_pb2 import Resource


################################################################################
"""
# NimbleBox.ai Serving

This is the proposed interface for the NimbleBox.ai Serving API. We want to keep
the highest levels of consistency with the NBX-Jobs API.
"""
################################################################################

@lru_cache()
def _get_deployment_data(name: str = "", id: str = "", *, workspace_id: str = "") -> Tuple[str, str]:
  """
  Get the deployment data, either by name or id

  Args:
    name (str, optional): Name of the deployment. Defaults to "".
    id (str, optional): ID of the deployment. Defaults to "".

  Returns:
    Tuple[str, str]: (id, name)
  """
  # print("Getting deployment data", name, id, workspace_id)
  if (not name and not id) or (name and id):
    logger.warning("Must provide either name or id")
    return None, None
  # filter and get "id" and "name"
  workspace_id = workspace_id or secret.workspace_id

  # get the deployment
  serving: Serving = mpb.rpc(
    nbox_serving_service_stub.GetServing,
    ServingRequest(
      serving=Serving(name=name, id=id),
      auth_info = auth_info_pb()
    ),
    "Could not get deployment",
    raise_on_error=True
  )

  return serving.id, serving.name

def print_serving_list(sort: str = "created_on", *, workspace_id: str = ""):
  """
  Print the list of deployments

  Args:
    sort (str, optional): Sort by. Defaults to "created_on".
  """
  def _get_time(t):
    return datetime.fromtimestamp(int(float(t))).strftime("%Y-%m-%d %H:%M:%S")

  workspace_id = workspace_id or secret.workspace_id
  all_deployments: ServingListResponse = mpb.rpc(
    nbox_serving_service_stub.ListServings,
    ServingListRequest(
      auth_info=auth_info_pb(),
      limit=10
    ),
    "Could not get deployments",
    raise_on_error=True
  )
  # sorted_depls = sorted(all_deployments, key = lambda x: x[sort], reverse = sort == "created_on")
  # headers = ["created_on", "id", "name", "pinned_id", "pinned_name", "pinned_last_updated"]
  # [TODO] add sort by create time
  headers = ["id", "name", "pinned_id", "pinned_name", "pinned_last_updated"]
  all_depls = []
  for depl in all_deployments.servings:
    _depl = (depl.id, depl.name)
    pinned = depl.models[0] if len(depl.models) else None
    if not pinned:
      _depl += (None, None,)
    else:
      _depl += (pinned.id, pinned.name, _get_time(pinned.created_at.seconds))
    all_depls.append(_depl)

  for l in tabulate.tabulate(all_depls, headers).splitlines():
    logger.info(l)


class Serve:
  # status = staticmethod(print_serving_list)
  # upload: 'Serve' = staticmethod(partial(upload_job_folder, "serving"))

  def __init__(self, serving_id: str = "", model_id: str = "", *, workspace_id: str = "") -> None:
    """Python wrapper for NBX-Serving gRPC API

    Args:
      serving_id (str, optional): Serving ID. Defaults to None.
      model_id (str, optional): Model ID. Defaults to None.
    """
    self.id = serving_id
    self.model_id = model_id
    self.workspace_id = workspace_id or secret.workspace_id
    if workspace_id is None:
      raise DeprecationWarning("Personal workspace does not support serving")
    else:
      serving_id, serving_name = _get_deployment_data(name = "", id = self.id, workspace_id = self.workspace_id) # TODO add name support
    self.serving_id = serving_id
    self.serving_name = serving_name
    self.ws_stub = nbox_ws_v1.deployments

  def __repr__(self) -> str:
    x = f"nbox.Serve('{self.id}', '{self.workspace_id}'"
    if self.model_id is not None:
      x += f", model_id = '{self.model_id}'"
    x += ")"
    return x

  # there are things that are on the serving group level
  
  def logs(self, f = sys.stdout):
    """Get the logs of the model deployment

    Args:
      f (file, optional): File to write the logs to. Defaults to sys.stdout.
    """
    logger.debug(f"Streaming logs of model '{self.model_id}'")
    for model_log in mpb.streaming_rpc(
      nbox_model_service_stub.ModelLogs,
      ModelRequest(
        model = ModelProto(
          id = self.model_id,
          serving_group_id = self.serving_id
        ),
        auth_info = auth_info_pb(),
      ),
      f"Could not get logs of model {self.model_id}, only live logs are available",
      False
    ):
      for log in model_log.log:
        f.write(log)
        f.flush()

  def change_serving_resources(self, resource: Dict[str, Any]):
    if type(resource, Resource):
      pass
    elif type(resource, dict):
      resource = mpb.dict_to_message(resource, Resource())
    else:
      raise TypeError(f"resource must be of type dict or Resource, not {type(resource)}")

    logger.info(f"Updating resources of a serving group: {self.serving_id}")
    mpb.rpc(
      nbox_serving_service_stub.UpdateServing,
      UpdateServingRequest(
        model=Serving(
          id=self.model_id,
          resource=resource
        ),
        update_mask=FieldMask(paths=["resource"]),
        auth_info = auth_info_pb()
      ),
      "Could not change resources of serving group",
      raise_on_error = True
    )
    pass

  # there are things on the model level

  def deploy(self, tag: str = "", feature_gates: Dict[str, str] = {}, resource: Dict[str, Any] = {}):
    if not self.model_id:
      raise ValueError("Model ID is required")

    if type(resource) ==  Resource:
      pass
    elif type(resource) ==  dict:
      resource = mpb.dict_to_message(resource, Resource())
    else:
      raise TypeError(f"resource must be of type dict or Resource, not {type(resource)}")

    model = ModelProto(
      id = self.model_id,
      serving_group_id = self.serving_id,
      resource = resource
    )
    if tag:
      model.feature_gates.update({"SetModelMetadata": tag})
    if feature_gates:
      model.feature_gates.update(feature_gates)
    logger.info(f"Deploying model {self.model_id} to deployment {self.serving_id} with tag: '{tag}' and feature gates: {feature_gates}")
    try:
      nbox_model_service_stub.Deploy(ModelRequest(model = model, auth_info = auth_info_pb()))
    except grpc.RpcError as e:
      logger.error(lo(
        f"Could not deploy model {self.model_id} to deployment {self.serving_id}\n",
        f"gRPC Code: {e.code()}\n"
        f"    Error: {e.details()}",
      ))

  def pin(self) -> bool:
    """Pin a model to the deployment

    Args:
      model_id (str, optional): Model ID. Defaults to None.
      workspace_id (str, optional): Workspace ID. Defaults to "".
    """
    if not self.model_id:
      raise ValueError("Model ID is required")

    logger.info(f"Pin model {self.model_id} to deployment {self.serving_id}")
    mpb.rpc(
      nbox_model_service_stub.SetModelPin,
      ModelRequest(
        model = ModelProto(
          id = self.model_id,
          serving_group_id = self.serving_id,
          pin_status = ModelProto.PinStatus.PIN_STATUS_PINNED
        ),
        auth_info = auth_info_pb()
      ),
      "Could not pin model",
      raise_on_error = True
    )
  
  def unpin(self) -> bool:
    """Pin a model to the deployment

    Args:
      model_id (str, optional): Model ID. Defaults to None.
      workspace_id (str, optional): Workspace ID. Defaults to "".
    """
    if not self.model_id:
      raise ValueError("Model ID is required")

    logger.info(f"Unpin model {self.model_id} to deployment {self.serving_id}")
    mpb.rpc(
      nbox_model_service_stub.SetModelPin,
      ModelRequest(
        model = ModelProto(
          id = self.model_id,
          serving_group_id = self.serving_id,
          pin_status = ModelProto.PinStatus.PIN_STATUS_UNPINNED
        ),
        auth_info = auth_info_pb(),
      ),
      "Could not unpin model",
      raise_on_error = True
    )
  
  def scale(self, replicas: int) -> bool:
    """Scale the model deployment

    Args:
      replicas (int): Number of replicas
    """
    if not self.model_id:
      raise ValueError("Model ID is required to scale a model. Pass with --model_id")
    if replicas < 0:
      raise ValueError("Replicas must be greater than or equal to 0")

    logger.info(f"Scale model deployment {self.model_id} to {replicas} replicas")
    mpb.rpc(
      nbox_model_service_stub.UpdateModel,
      UpdateModelRequest(
        model=ModelProto(
          id=self.model_id,
          serving_group_id=self.serving_id,
          replicas=replicas
        ),
        update_mask=FieldMask(paths=["replicas"]),
        auth_info = auth_info_pb()
      ),
      "Could not scale deployment",
      raise_on_error = True
    )

  def change_model_resources(self, resource: Dict[str, Any]):
    if type(resource) ==  Resource:
      pass
    elif type(resource) ==  dict:
      resource = mpb.dict_to_message(resource, Resource())
    else:
      raise TypeError(f"resource must be of type dict or Resource, not {type(resource)}")

    logger.warn(f"Updating resources of a running model '{self.model_id}', this will result in a restart!")
    mpb.rpc(
      nbox_model_service_stub.UpdateModel,
      UpdateModelRequest(
        model=ModelProto(
          id=self.model_id,
          serving_group_id=self.serving_id,
          resource=resource
        ),
        update_mask=FieldMask(paths=["resource"]),
        auth_info = auth_info_pb()
      ),
      "Could not change resources of a running model",
      raise_on_error = True
    )


from functools import partial

from nbox.jd_core.upload import upload_job_folder
from nbox.jd_core.jobs import get_job_list, Job
from nbox.jd_core.serving import print_serving_list, Serve

JobsCli = {
  "status": get_job_list,
  "upload": partial(upload_job_folder, "job")
}


ServeCli = {
  "status": print_serving_list,
  "upload": partial(upload_job_folder, "serving")
}



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import collections.abc
import common.common_pb2
import google.protobuf.empty_pb2
import google.protobuf.wrappers_pb2
import grpc
import jobs.job_pb2
import jobs.nbox_ws_pb2

class WSJobServiceStub:
    """
    WSJobService is used by nbox to interface with Web Server. WSJobService
    manages DB and NBX Platform Information. All compute resource related actions
    are fowarded to WSJobsService running on AWS.

    // Generated according to
    https://cloud.google.com/apis/design/standard_methods
    """

    def __init__(self, channel: grpc.Channel) -> None: ...
    GetJob: grpc.UnaryUnaryMultiCallable[
        jobs.nbox_ws_pb2.JobRequest,
        jobs.job_pb2.Job,
    ]
    """GetJob returns the job with the given info"""
    ListJobs: grpc.UnaryUnaryMultiCallable[
        jobs.nbox_ws_pb2.ListJobsRequest,
        jobs.nbox_ws_pb2.ListJobsResponse,
    ]
    """ListJobs returns a list of jobs for the given workspace/user"""
    UploadJobCode: grpc.UnaryUnaryMultiCallable[
        jobs.nbox_ws_pb2.JobRequest,
        jobs.job_pb2.Job,
    ]
    """To upload Job Code to S3"""
    CreateJob: grpc.UnaryUnaryMultiCallable[
        jobs.nbox_ws_pb2.JobRequest,
        jobs.job_pb2.Job,
    ]
    """CreateJob creates a new job"""
    UpdateJob: grpc.UnaryUnaryMultiCallable[
        jobs.nbox_ws_pb2.UpdateJobRequest,
        jobs.job_pb2.Job,
    ]
    """UpdateJob updates the given job"""
    DeleteJob: grpc.UnaryUnaryMultiCallable[
        jobs.nbox_ws_pb2.JobRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """DeleteJob deletes the given job"""
    GetJobLogs: grpc.UnaryStreamMultiCallable[
        jobs.nbox_ws_pb2.JobRequest,
        common.common_pb2.Log,
    ]
    """GetJobLogs returns logs from given job"""
    TriggerJob: grpc.UnaryUnaryMultiCallable[
        jobs.nbox_ws_pb2.JobRequest,
        google.protobuf.wrappers_pb2.BoolValue,
    ]
    """TriggerJob triggers the given job"""
    UpdateRun: grpc.UnaryUnaryMultiCallable[
        jobs.nbox_ws_pb2.UpdateRunRequest,
        google.protobuf.empty_pb2.Empty,
    ]
    """UpdateRun update the run for given job"""
    GetRunToken: grpc.UnaryUnaryMultiCallable[
        jobs.nbox_ws_pb2.JobRequest,
        google.protobuf.wrappers_pb2.StringValue,
    ]
    """GetRunToken returns a tokn to be used for `/UpdateRun`"""

class WSJobServiceServicer(metaclass=abc.ABCMeta):
    """
    WSJobService is used by nbox to interface with Web Server. WSJobService
    manages DB and NBX Platform Information. All compute resource related actions
    are fowarded to WSJobsService running on AWS.

    // Generated according to
    https://cloud.google.com/apis/design/standard_methods
    """

    @abc.abstractmethod
    def GetJob(
        self,
        request: jobs.nbox_ws_pb2.JobRequest,
        context: grpc.ServicerContext,
    ) -> jobs.job_pb2.Job:
        """GetJob returns the job with the given info"""
    @abc.abstractmethod
    def ListJobs(
        self,
        request: jobs.nbox_ws_pb2.ListJobsRequest,
        context: grpc.ServicerContext,
    ) -> jobs.nbox_ws_pb2.ListJobsResponse:
        """ListJobs returns a list of jobs for the given workspace/user"""
    @abc.abstractmethod
    def UploadJobCode(
        self,
        request: jobs.nbox_ws_pb2.JobRequest,
        context: grpc.ServicerContext,
    ) -> jobs.job_pb2.Job:
        """To upload Job Code to S3"""
    @abc.abstractmethod
    def CreateJob(
        self,
        request: jobs.nbox_ws_pb2.JobRequest,
        context: grpc.ServicerContext,
    ) -> jobs.job_pb2.Job:
        """CreateJob creates a new job"""
    @abc.abstractmethod
    def UpdateJob(
        self,
        request: jobs.nbox_ws_pb2.UpdateJobRequest,
        context: grpc.ServicerContext,
    ) -> jobs.job_pb2.Job:
        """UpdateJob updates the given job"""
    @abc.abstractmethod
    def DeleteJob(
        self,
        request: jobs.nbox_ws_pb2.JobRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """DeleteJob deletes the given job"""
    @abc.abstractmethod
    def GetJobLogs(
        self,
        request: jobs.nbox_ws_pb2.JobRequest,
        context: grpc.ServicerContext,
    ) -> collections.abc.Iterator[common.common_pb2.Log]:
        """GetJobLogs returns logs from given job"""
    @abc.abstractmethod
    def TriggerJob(
        self,
        request: jobs.nbox_ws_pb2.JobRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.wrappers_pb2.BoolValue:
        """TriggerJob triggers the given job"""
    @abc.abstractmethod
    def UpdateRun(
        self,
        request: jobs.nbox_ws_pb2.UpdateRunRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.empty_pb2.Empty:
        """UpdateRun update the run for given job"""
    @abc.abstractmethod
    def GetRunToken(
        self,
        request: jobs.nbox_ws_pb2.JobRequest,
        context: grpc.ServicerContext,
    ) -> google.protobuf.wrappers_pb2.StringValue:
        """GetRunToken returns a tokn to be used for `/UpdateRun`"""

def add_WSJobServiceServicer_to_server(servicer: WSJobServiceServicer, server: grpc.Server) -> None: ...



from datetime import datetime, timezone, timedelta
from google.protobuf.timestamp_pb2 import Timestamp

from nbox.hyperloop.jobs.job_pb2 import Job as JobProto

class Schedule:
  def __init__(
    self,
    hour: int  = None,
    minute: int = None,
    days: list = [],
    months: list = [],
    starts: datetime = None,
    ends: datetime = None,
  ):
    """Make scheduling natural. Uses 24-hour nomenclature.

    Args:
      hour (int): Hour of the day, if only this value is passed it will run every `hour`
      minute (int): Minute of the hour, if only this value is passed it will run every `minute`
      days (str/list, optional): List of days (first three chars) of the week, if not passed it will run every day.
      months (str/list, optional): List of months (first three chars) of the year, if not passed it will run every month.
      starts (datetime, optional): UTC Start time of the schedule, if not passed it will start now.
      ends (datetime, optional): UTC End time of the schedule, if not passed it will end in 7 days.

    Examples:

      # 4:20PM everyday
      Schedule(16, 20)

      # 4:20AM every friday
      Schedule(4, 20, ["fri"])

      # 4:20AM every friday from jan to feb
      Schedule(4, 20, ["fri"], ["jan", "feb"])

      # 4:20PM everyday starting in 2 days and runs for 3 days
      starts = datetime.now(timezone.utc) + timedelta(days = 2) # NOTE: that time is in UTC
      Schedule(16, 20, starts = starts, ends = starts + timedelta(days = 3))

      # Every 1 hour
      Schedule(1)

      # Every 69 minutes
      Schedule(minute = 69)
    """
    self.hour = hour
    self.minute = minute

    self._is_repeating = self.hour or self.minute
    self.mode = None
    if self.hour == None and self.minute == None:
      raise ValueError("Atleast one of hour or minute should be passed")
    elif self.hour != None and self.minute != None:
      assert self.hour in list(range(0, 24)), f"Hour must be in range 0-23, got {self.hour}"
      assert self.minute in list(range(0, 60)), f"Minute must be in range 0-59, got {self.minute}"
    elif self.hour != None:
      assert self.hour in list(range(0, 24)), f"Hour must be in range 0-23, got {self.hour}"
      self.mode = "every"
      self.minute = datetime.now(timezone.utc).strftime("%m") # run every this minute past this hour
      self.hour = f"*/{self.hour}"
    elif self.minute != None:
      self.hour = self.minute // 60
      assert self.hour in list(range(0, 24)), f"Hour must be in range 0-23, got {self.hour}"
      self.minute = f"*/{self.minute % 60}"
      self.hour = f"*/{self.hour}" if self.hour > 0 else "*"
      self.mode = "every"

    _days = {k:str(i) for i,k in enumerate(["sun","mon","tue","wed","thu","fri","sat"])}
    _months = {k:str(i+1) for i,k in enumerate(["jan","feb","mar","apr","may","jun","jul","aug","sep","oct","nov","dec"])}

    if isinstance(days, str):
      days = [days]
    if isinstance(months, str):
      months = [months]

    diff = set(days) - set(_days.keys())
    if len(diff):
      raise ValueError(f"Invalid days: {diff}")
    self.days = ",".join([_days[d] for d in days]) if days else "*"

    diff = set(months) - set(_months.keys())
    if len(diff):
      raise ValueError(f"Invalid months: {diff}")
    self.months = ",".join([_months[m] for m in months]) if months else "*"

    self.starts = starts or datetime.now(timezone.utc)
    self.ends = ends or datetime.now(timezone.utc) + timedelta(days = 7)

  @property
  def cron(self):
    """Get the cron string for the given schedule"""
    if self.mode == "every":
      return f"{self.minute} {self.hour} * * *"
    return f"{self.minute} {self.hour} * {self.months} {self.days}"

  def get_dict(self):
    """Get the dictionary representation of this Schedule"""
    return {"cron": self.cron, "mode": self.mode, "starts": self.starts, "ends": self.ends}

  def get_message(self) -> JobProto.Schedule:
    """Get the JobProto.Schedule object for this Schedule"""
    _starts = Timestamp()
    _starts.GetCurrentTime()
    _ends = Timestamp()
    _ends.FromDatetime(self.ends)
    return JobProto.Schedule(
      start = _starts,
      end = _ends,
      cron = self.cron
    )

  def __repr__(self):
    return str(self.get_dict())


from nbox.jd_core.schedule import Schedule
from nbox.jd_core.jobs import Job, _get_job_data, get_job_list
from nbox.jd_core.serving import Serve, _get_deployment_data, print_serving_list
from nbox.jd_core.upload import upload_job_folder
from nbox.jd_core.cli import JobsCli, ServeCli



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import nbox.relics.proto.relics_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _Actions:
    ValueType = typing.NewType('ValueType', builtins.int)
    V: typing_extensions.TypeAlias = ValueType
class _ActionsEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Actions.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    NONE: _Actions.ValueType  # 0
    CREATE: _Actions.ValueType  # 1
    UPDATE: _Actions.ValueType  # 2
    DELETE: _Actions.ValueType  # 3
class Actions(_Actions, metaclass=_ActionsEnumTypeWrapper):
    """This is a set of generic RPCs which can be used to mark operation type for relics"""
    pass

NONE: Actions.ValueType  # 0
CREATE: Actions.ValueType  # 1
UPDATE: Actions.ValueType  # 2
DELETE: Actions.ValueType  # 3
global___Actions = Actions


class CreateRelicRequest(google.protobuf.message.Message):
    """things related to general WebAPIs"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    REGION_FIELD_NUMBER: builtins.int
    AUTH_FIELD_NUMBER: builtins.int
    BUCKET_META_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the workspace id for which we are creating the relic"""

    name: typing.Text
    region: typing.Text
    @property
    def auth(self) -> proto.relics_pb2.BackendInfo: ...
    @property
    def bucket_meta(self) -> proto.relics_pb2.BucketMetadata:
        """If the bucket is in their platform then they tell us that they want one there
        else we are going to assume that they want one in our platform
        """
        pass
    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        name: typing.Text = ...,
        region: typing.Text = ...,
        auth: typing.Optional[proto.relics_pb2.BackendInfo] = ...,
        bucket_meta: typing.Optional[proto.relics_pb2.BucketMetadata] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth",b"auth","bucket_meta",b"bucket_meta"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth",b"auth","bucket_meta",b"bucket_meta","name",b"name","region",b"region","workspace_id",b"workspace_id"]) -> None: ...
global___CreateRelicRequest = CreateRelicRequest

class ListRelicsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    RELIC_NAME_FIELD_NUMBER: builtins.int
    PAGE_NO_FIELD_NUMBER: builtins.int
    RELIC_ID_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the workspace id for which we are creating the relic"""

    relic_name: typing.Text
    """the name of the relic"""

    page_no: builtins.int
    """the page number for pagination needs"""

    relic_id: typing.Text
    """the ID of the parent relic"""

    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        relic_name: typing.Text = ...,
        page_no: builtins.int = ...,
        relic_id: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["page_no",b"page_no","relic_id",b"relic_id","relic_name",b"relic_name","workspace_id",b"workspace_id"]) -> None: ...
global___ListRelicsRequest = ListRelicsRequest

class ListRelicsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    RELICS_FIELD_NUMBER: builtins.int
    TOTAL_RELICS_FIELD_NUMBER: builtins.int
    @property
    def relics(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[proto.relics_pb2.Relic]: ...
    total_relics: builtins.int
    """the total number of relics used for pagination"""

    def __init__(self,
        *,
        relics: typing.Optional[typing.Iterable[proto.relics_pb2.Relic]] = ...,
        total_relics: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["relics",b"relics","total_relics",b"total_relics"]) -> None: ...
global___ListRelicsResponse = ListRelicsResponse

class ListRelicFilesRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    RELIC_ID_FIELD_NUMBER: builtins.int
    RELIC_NAME_FIELD_NUMBER: builtins.int
    PREFIX_FIELD_NUMBER: builtins.int
    FILE_NAME_FIELD_NUMBER: builtins.int
    PAGE_NO_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the workspace id for which we are creating the relic"""

    relic_id: typing.Text
    """the id of the relic"""

    relic_name: typing.Text
    """the id of the relic"""

    prefix: typing.Text
    """the prefix for the files"""

    file_name: typing.Text
    """for searching through the files, we may not support this functionality
    right now, but this is a crucial system to have
    """

    page_no: builtins.int
    """the page number for pagination needs"""

    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        relic_id: typing.Text = ...,
        relic_name: typing.Text = ...,
        prefix: typing.Text = ...,
        file_name: typing.Text = ...,
        page_no: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["file_name",b"file_name","page_no",b"page_no","prefix",b"prefix","relic_id",b"relic_id","relic_name",b"relic_name","workspace_id",b"workspace_id"]) -> None: ...
global___ListRelicFilesRequest = ListRelicFilesRequest

class ListRelicFilesResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    FILES_FIELD_NUMBER: builtins.int
    TOTAL_FILES_FIELD_NUMBER: builtins.int
    @property
    def files(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[proto.relics_pb2.RelicFile]: ...
    total_files: builtins.int
    """the total number of files used for pagination"""

    def __init__(self,
        *,
        files: typing.Optional[typing.Iterable[proto.relics_pb2.RelicFile]] = ...,
        total_files: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["files",b"files","total_files",b"total_files"]) -> None: ...
global___ListRelicFilesResponse = ListRelicFilesResponse

class ActivityLogRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    RELIC_ID_FIELD_NUMBER: builtins.int
    RELIC_NAME_FIELD_NUMBER: builtins.int
    USERNAME_FIELD_NUMBER: builtins.int
    FROM_TIMESTAMP_NS_FIELD_NUMBER: builtins.int
    TO_TIMESTAMP_NS_FIELD_NUMBER: builtins.int
    PAGE_NO_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the workspace id for which we are creating the relic"""

    relic_id: typing.Text
    """the id of the relic"""

    relic_name: typing.Text
    """the id of the relic"""

    username: typing.Text
    """the username for which we want to get the logs"""

    from_timestamp_ns: builtins.int
    """the timestamp from which we want to get the logs"""

    to_timestamp_ns: builtins.int
    """the timestamp to which we want to get the logs"""

    page_no: builtins.int
    """the page number for pagination needs"""

    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        relic_id: typing.Text = ...,
        relic_name: typing.Text = ...,
        username: typing.Text = ...,
        from_timestamp_ns: builtins.int = ...,
        to_timestamp_ns: builtins.int = ...,
        page_no: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["from_timestamp_ns",b"from_timestamp_ns","page_no",b"page_no","relic_id",b"relic_id","relic_name",b"relic_name","to_timestamp_ns",b"to_timestamp_ns","username",b"username","workspace_id",b"workspace_id"]) -> None: ...
global___ActivityLogRequest = ActivityLogRequest

class ActivityLog(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    TIMESTAMP_NS_FIELD_NUMBER: builtins.int
    RELIC_ID_FIELD_NUMBER: builtins.int
    OPERATION_FIELD_NUMBER: builtins.int
    OBJECT_SIZE_FIELD_NUMBER: builtins.int
    USERNAME_FIELD_NUMBER: builtins.int
    OBJECT_NAME_FIELD_NUMBER: builtins.int
    timestamp_ns: builtins.int
    """the timestamp at which the logs were created"""

    relic_id: typing.Text
    """the id of the relic"""

    operation: global___Actions.ValueType
    """the operation that was performed"""

    object_size: builtins.float
    """the size of the object that was operated on"""

    username: typing.Text
    """the username of the user who performed the operation"""

    object_name: typing.Text
    """the name of the object that was operated on"""

    def __init__(self,
        *,
        timestamp_ns: builtins.int = ...,
        relic_id: typing.Text = ...,
        operation: global___Actions.ValueType = ...,
        object_size: builtins.float = ...,
        username: typing.Text = ...,
        object_name: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["object_name",b"object_name","object_size",b"object_size","operation",b"operation","relic_id",b"relic_id","timestamp_ns",b"timestamp_ns","username",b"username"]) -> None: ...
global___ActivityLog = ActivityLog

class ActivityLogResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    LOGS_FIELD_NUMBER: builtins.int
    @property
    def logs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ActivityLog]: ...
    def __init__(self,
        *,
        logs: typing.Optional[typing.Iterable[global___ActivityLog]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["logs",b"logs"]) -> None: ...
global___ActivityLogResponse = ActivityLogResponse

class ListBucketRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    AUTH_FIELD_NUMBER: builtins.int
    workspace_id: typing.Text
    """the workspace id for which we are creating the relic"""

    @property
    def auth(self) -> proto.relics_pb2.BackendInfo: ...
    def __init__(self,
        *,
        workspace_id: typing.Text = ...,
        auth: typing.Optional[proto.relics_pb2.BackendInfo] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth",b"auth"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth",b"auth","workspace_id",b"workspace_id"]) -> None: ...
global___ListBucketRequest = ListBucketRequest

class ListBucketResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    BUCKETS_FIELD_NUMBER: builtins.int
    @property
    def buckets(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]: ...
    def __init__(self,
        *,
        buckets: typing.Optional[typing.Iterable[typing.Text]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["buckets",b"buckets"]) -> None: ...
global___ListBucketResponse = ListBucketResponse


"""
With NimbleBox you can run multi-cluster wide workloads from anywhere. This requires capabilities around distributed computing,
process management. The code here is tested along with `nbox.Relic` to perform distributed local and cloud processing.

{% CallOut variant="success" label="If you find yourself using this reach out to NimbleBox support." /%}
"""

import os
import json

import nbox.utils as U
from nbox.utils import logger, lo
from nbox.relics import Relics
from nbox.operator import Operator
from nbox.auth import secret, AuthConfig
from nbox.nbxlib.tracer import Tracer
from nbox.hyperloop.jobs.job_pb2 import Job
from nbox.nbxlib.serving import serve_operator

from nbox.lmao import ExperimentConfig, LiveConfig, LMAO_RM_PREFIX, LMAO_SERVING_FILE
from nbox.projects import Project, ProjectState

# Manager
class LocalNBXLet(Operator):
  def __init__(self, op: Operator, in_key: str, out_key: str):
    super().__init__()
    self.op = op
    self.in_key = in_key
    self.out_key = out_key

  def __repr__(self):
    return f"LocalNBXLet({self.op.__qualname__}, {self.in_key}, {self.out_key})"

  def forward(self):
    x = U.from_pickle(self.in_key)
    y = self.op(*x)
    U.to_pickle(y, self.out_key)


class NBXLet(Operator):
  def __init__(self, op: Operator):
    """The Operator that runs the things on any pod on the NimbleBox Jobs + Deploy platform.
    Name a parody of kubelet, dockerlet, raylet, etc"""
    super().__init__()
    self.op = op

  def run(self):
    """Run this as a batch process"""
    # get the user defined tag 
    run_tag = os.getenv("NBOX_RUN_METADATA", "")
    logger.info(f"Run Tag: {run_tag}")

    status = Job.Status.ERROR
    try:
      tracer = Tracer()
      if hasattr(self.op._tracer, "job_proto"):
        self.op.thaw(self.op._tracer.job_proto)
      workspace_id = tracer.workspace_id
      logger.info(f"Workspace Id: {workspace_id}")

      # this is important since nbox uses AuthConfig.workspace_id place to get workspace_id from while the init_container
      # might place it at a different place. as of this writing, init_container -> "workspace_id" and nbox -> "config.global.workspace_id"
      secret.put(AuthConfig.workspace_id, workspace_id, True) 

      job_id = tracer.job_id
      self.op.propagate(_tracer = tracer)

      # in the NimbleBox system we provide tags for each key which essentially tells what is the behaviour
      # of the job. For example if it contains the string LMAO which means we need to initialise a couple
      # of things, or this can be any other job type
      args, kwargs = (), {}
      if run_tag.startswith(LMAO_RM_PREFIX):
        # originally we had a strategy to use Relics to store the information about the initialisation and passed args
        # however we are not removing that because we don't want to spend access money when we are anyways storing all
        # the information in the LMAO DB. so now we get the details of the run and get all the information from there.
        #
        # update (27/02/23): We completed the integration of LMAO with NBX-Projects to get beautifully simple interface
        #   for your entire MLOps pipeline. So copying the style from LMAO, we have something called ProjectState that
        #   has some variables that simplify the client side code when they don't have to pass any ids, it's all
        #   inferred.

        project_id, exp_id = run_tag[len(LMAO_RM_PREFIX):].split("/")
        logger.info(f"Project name (Experiment ID): {project_id} ({exp_id})")
        ProjectState.project_id = project_id
        ProjectState.experiment_id = exp_id

        # create the central project class and get the experiment tracker
        proj = Project()
        logger.info(lo("Project data:", **proj.data))
        exp_tracker = proj.get_exp_tracker()
        lmao_run = exp_tracker.run
        exp_config = ExperimentConfig.from_json(lmao_run.config)
        kwargs = exp_config.run_kwargs

      elif run_tag.startswith(RAW_DIST_RM_PREFIX):
        relic = Relics(RAW_DIST_RELIC_NAME, workspace_id)
        _pkl_id_in = run_tag[len(RAW_DIST_RM_PREFIX):] + "_in"
        logger.info(f"Looking for init.pkl at {_pkl_id_in}")
        (args, kwargs) = relic.get_object(_pkl_id_in)

      elif run_tag.startswith(SILK_RM_PREFIX):
        # 27/03/2023: We are adding a new job type called Silk
        trace_id = run_tag[len(SILK_RM_PREFIX):]
        kwargs = {"trace_id": trace_id}

      # call the damn thing
      st = U.SimplerTimes.get_now_i64()
      logger.debug(f"Args: {args}\nKwargs: {kwargs}")
      out = self.op(*args, **kwargs)

      # save the output to the relevant place, LMAO jobs are not saved to the relic
      time_taken = U.SimplerTimes.get_now_i64() - st
      logger.info(lo(
        f"NBXLet: {run_tag} job completed, here's some stats:",
        time_taken = time_taken
      ))

      if run_tag.startswith(RAW_DIST_RM_PREFIX):
        _pkl_id_out = run_tag[len(RAW_DIST_RM_PREFIX):] + "_out"
        logger.info(f"Storing for output at {_pkl_id_out}")
        relic.put_object(_pkl_id_out, out)

      # last step mark as completed
      status = Job.Status.COMPLETED
    except Exception as e:
      U.log_traceback()
    finally:
      logger.info(f"Job {job_id} completed with status {status}")
      if hasattr(tracer, "job_proto"):
        tracer.job_proto.status = status
        tracer._rpc(f"RPC error in ending job {job_id}")
      U.hard_exit_program()

  def serve(self, **serve_kwargs):
    """Run a serving API endpoint"""
    run_tag = os.getenv("NBOX_MODEL_METADATA", "")
    logger.info(f"Run Tag: {run_tag}")

    try:
      if run_tag.startswith(LMAO_RM_PREFIX):
        project_id, tracker_id = run_tag[len(LMAO_RM_PREFIX):].split("/")
        logger.info(f"Project name (Tracker ID): {project_id} ({tracker_id})")
        ProjectState.project_id = project_id
        ProjectState.serving_id = tracker_id

        # create the central project class and get the experiment tracker
        proj = Project()
        logger.info(lo("Project data:", **proj.data))
        # live_tracker = proj.get_live_tracker()
        # tracker_config = LiveConfig.from_json(live_tracker.serving.config)

      # now start serving
      serve_operator(op_or_app = self.op, **serve_kwargs)
    except Exception as e:
      U.log_traceback()
      logger.error(f"Failed to serve operator: {e}")
      U.hard_exit_program()

# Here's all the different tags that we use
# 
# Raw dist system uses Relics as an intermediate storage and jobs as compute nodes
# this is not the most optimal way but this is technology demonstration
RAW_DIST_RELIC_NAME = "tmp_cache"
RAW_DIST_RM_PREFIX = "NBXOperatorRawDist-"
RAW_DIST_ENV_VAR_PREFIX = "NBX_OperatorRawDist_"

# Silk is a production grade pipeline execution engine that can run any python code
SILK_RM_PREFIX = "ComputeSilk/"



"""
`nbox.Job` and `nbox.Serve` are wrappers to the NBX-Jobs and NBX-Deploy APIs and contains staticmethods for convinience from the CLI.

* `datetime.now(timezone.utc)` is incorrect, use [this](https://blog.ganssle.io/articles/2019/11/utcnow.html) method.
"""

import os
import sys
import grpc
import tabulate
from typing import Tuple, List, Dict, Any
from functools import lru_cache, partial
from datetime import datetime, timedelta, timezone
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.field_mask_pb2 import FieldMask

import nbox.utils as U
from nbox.auth import secret, AuthConfig, auth_info_pb
from nbox.utils import logger, lo
from nbox.version import __version__
from nbox import messages as mpb
# from nbox.messages import rpc, streaming_rpc
from nbox.init import nbox_grpc_stub, nbox_ws_v1, nbox_serving_service_stub, nbox_model_service_stub
from nbox.nbxlib.astea import Astea, IndexTypes as IT

from nbox.hyperloop.jobs.nbox_ws_pb2 import JobRequest
from nbox.hyperloop.jobs.job_pb2 import Job as JobProto
from nbox.hyperloop.jobs.dag_pb2 import DAG as DAGProto
from nbox.hyperloop.common.common_pb2 import Resource
from nbox.hyperloop.jobs.nbox_ws_pb2 import (
  ListJobsRequest,
  ListJobsResponse,
  UpdateJobRequest
)
from nbox.hyperloop.deploy.serve_pb2 import (
  ServingListResponse,
  ServingRequest,
  Serving,
  ServingListRequest,
  UpdateServingRequest,
  ModelRequest,
  Model as ModelProto,
  UpdateModelRequest,
)


DEFAULT_RESOURCE = Resource(
  cpu = "128m",         # 100mCPU
  memory = "256Mi",     # MiB
  disk_size = "3Gi",    # GiB
  gpu = "none",         # keep "none" for no GPU
  gpu_count = "0",      # keep "0" when no GPU
  timeout = 120_000,    # 2 minutes between attempts
  max_retries = 2,      # third times the charm :P
)

from nbox.jd_core import (
    Schedule,
    _get_job_data,
    get_job_list,
    Job,
    _get_deployment_data,
    print_serving_list,
    Serve,
    JobsCli,
    ServeCli
)


# class Schedule:
#   def __init__(
#     self,
#     hour: int  = None,
#     minute: int = None,
#     days: list = [],
#     months: list = [],
#     starts: datetime = None,
#     ends: datetime = None,
#   ):
#     """Make scheduling natural. Uses 24-hour nomenclature.

#     Args:
#       hour (int): Hour of the day, if only this value is passed it will run every `hour`
#       minute (int): Minute of the hour, if only this value is passed it will run every `minute`
#       days (str/list, optional): List of days (first three chars) of the week, if not passed it will run every day.
#       months (str/list, optional): List of months (first three chars) of the year, if not passed it will run every month.
#       starts (datetime, optional): UTC Start time of the schedule, if not passed it will start now.
#       ends (datetime, optional): UTC End time of the schedule, if not passed it will end in 7 days.

#     Examples:

#       # 4:20PM everyday
#       Schedule(16, 20)

#       # 4:20AM every friday
#       Schedule(4, 20, ["fri"])

#       # 4:20AM every friday from jan to feb
#       Schedule(4, 20, ["fri"], ["jan", "feb"])

#       # 4:20PM everyday starting in 2 days and runs for 3 days
#       starts = datetime.now(timezone.utc) + timedelta(days = 2) # NOTE: that time is in UTC
#       Schedule(16, 20, starts = starts, ends = starts + timedelta(days = 3))

#       # Every 1 hour
#       Schedule(1)

#       # Every 69 minutes
#       Schedule(minute = 69)
#     """
#     self.hour = hour
#     self.minute = minute

#     self._is_repeating = self.hour or self.minute
#     self.mode = None
#     if self.hour == None and self.minute == None:
#       raise ValueError("Atleast one of hour or minute should be passed")
#     elif self.hour != None and self.minute != None:
#       assert self.hour in list(range(0, 24)), f"Hour must be in range 0-23, got {self.hour}"
#       assert self.minute in list(range(0, 60)), f"Minute must be in range 0-59, got {self.minute}"
#     elif self.hour != None:
#       assert self.hour in list(range(0, 24)), f"Hour must be in range 0-23, got {self.hour}"
#       self.mode = "every"
#       self.minute = datetime.now(timezone.utc).strftime("%m") # run every this minute past this hour
#       self.hour = f"*/{self.hour}"
#     elif self.minute != None:
#       self.hour = self.minute // 60
#       assert self.hour in list(range(0, 24)), f"Hour must be in range 0-23, got {self.hour}"
#       self.minute = f"*/{self.minute % 60}"
#       self.hour = f"*/{self.hour}" if self.hour > 0 else "*"
#       self.mode = "every"

#     _days = {k:str(i) for i,k in enumerate(["sun","mon","tue","wed","thu","fri","sat"])}
#     _months = {k:str(i+1) for i,k in enumerate(["jan","feb","mar","apr","may","jun","jul","aug","sep","oct","nov","dec"])}

#     if isinstance(days, str):
#       days = [days]
#     if isinstance(months, str):
#       months = [months]

#     diff = set(days) - set(_days.keys())
#     if len(diff):
#       raise ValueError(f"Invalid days: {diff}")
#     self.days = ",".join([_days[d] for d in days]) if days else "*"

#     diff = set(months) - set(_months.keys())
#     if len(diff):
#       raise ValueError(f"Invalid months: {diff}")
#     self.months = ",".join([_months[m] for m in months]) if months else "*"

#     self.starts = starts or datetime.now(timezone.utc)
#     self.ends = ends or datetime.now(timezone.utc) + timedelta(days = 7)

#   @property
#   def cron(self):
#     """Get the cron string for the given schedule"""
#     if self.mode == "every":
#       return f"{self.minute} {self.hour} * * *"
#     return f"{self.minute} {self.hour} * {self.months} {self.days}"

#   def get_dict(self):
#     """Get the dictionary representation of this Schedule"""
#     return {"cron": self.cron, "mode": self.mode, "starts": self.starts, "ends": self.ends}

#   def get_message(self) -> JobProto.Schedule:
#     """Get the JobProto.Schedule object for this Schedule"""
#     _starts = Timestamp()
#     _starts.GetCurrentTime()
#     _ends = Timestamp()
#     _ends.FromDatetime(self.ends)
#     return JobProto.Schedule(
#       start = _starts,
#       end = _ends,
#       cron = self.cron
#     )

#   def __repr__(self):
#     return str(self.get_dict())



# ################################################################################
# """
# # Common Functions

# These functions are common to both NBX-Jobs and NBX-Deploy.
# """
# ################################################################################

# def upload_job_folder(
#   method: str,
#   init_folder: str,
#   id: str = "",
#   project_id: str = "",

#   # job / deploy rpc things
#   trigger: bool = False,
#   deploy: bool = True,
#   pin: bool = False,
#   feature_gates: dict = {},

#   # all the things for resources
#   resource_cpu: str = "",
#   resource_memory: str = "",
#   resource_disk_size: str = "",
#   resource_gpu: str = "",
#   resource_gpu_count: str = "",
#   resource_timeout: int = 0,
#   resource_max_retries: int = 0,

#   # deployment specific
#   model_name: str = "",

#   # X-type
#   serving_type: str = "nbox",

#   # there's no more need to pass the workspace_id anymore
#   workspace_id: str = "",

#   # some extra things for functionality
#   _ret: bool = False,

#   # finally everything else is assumed to be passed to the initialisation script
#   **init_kwargs
# ):
#   """Upload the code for a job or serving to the NBX.

#   ### Engineer's Note

#   This function is supposed to be exposed via CLI, you can of course make a programtic call to this as well. This is a reason
#   why this function can keep on taking many arguments. However with greatm number of arguments comes great responsibility, ie.
#   lots of if/else conditions. Broadly speaking this should manage all the complexity and pass along simple reduced intstructions
#   to the underlying methods. Currently the arguments that are required in Jinja templates are packed as `exe_jinja_kwargs` and
#   we call the `deploy_job` and `deploy_serving`.

#   Args:
#     method (str): The method to use, either "job" or "serving"
#     init_folder (str): folder with all the relevant files or ``file_path:fn_name`` pair so you can use it as the entrypoint.
#     name (str, optional): Name of the job. Defaults to "".
#     id (str, optional): ID of the job. Defaults to "".
#     project_id (str, optional): Project ID, if None uses the one from config. Defaults to "".
#     trigger (bool, optional): If uploading a "job" trigger the job after uploading. Defaults to False.
#     resource_cpu (str, optional): CPU resource. Defaults to "100m".
#     resource_memory (str, optional): Memory resource. Defaults to "128Mi".
#     resource_disk_size (str, optional): Disk size resource. Defaults to "3Gi".
#     resource_gpu (str, optional): GPU resource. Defaults to "none".
#     resource_gpu_count (str, optional): GPU count resource. Defaults to "0".
#     resource_timeout (int, optional): Timeout resource. Defaults to 120_000.
#     resource_max_retries (int, optional): Max retries resource. Defaults to 2.
#     cron (str, optional): Cron string for scheduling. Defaults to "".
#     workspace_id (str, optional): Workspace ID, if None uses the one from config. Defaults to "".
#     init_kwargs (dict): kwargs to pass to the `init` function / class, if possible
#   """
#   from nbox.network import deploy_job, deploy_serving
#   import nbox.nbxlib.operator_spec as ospec
#   from nbox.nbxlib.serving import SupportedServingTypes as SST
#   from nbox.projects import Project
  
#   OT = ospec.OperatorType

#   if method not in OT._valid_deployment_types():
#     raise ValueError(f"Invalid method: {method}, should be either {OT._valid_deployment_types()}")
#   if trigger and method != OT.JOB:
#     raise ValueError(f"Trigger can only be used with method='{OT.JOB}'")
#   if pin and method != OT.SERVING:
#     raise ValueError(f"Deploy and Pin can only be used with method='{OT.SERVING}'")
#   if model_name and method != OT.SERVING:
#     raise ValueError(f"model_name can only be used with '{OT.SERVING}'")
  
#   # get the correct ID based on the project_id
#   if (not project_id and not id) or (project_id and id):
#     raise ValueError("Either --project-id or --id must be present")
#   if project_id:
#     p = Project(project_id)
#     if method == OT.JOB:
#       id = p.get_job_id()
#     else:
#       id = p.get_deployment_id()
#     logger.info(f"Using project_id: {project_id}, found id: {id}")

#   if ":" not in init_folder:
#     # this means we are uploading a traditonal folder that contains a `nbx_user.py` file
#     # in this case the module is loaded on the local machine and so user will need to have
#     # everything installed locally. This was a legacy method before 0.10.0
#     logger.error(
#       'Old method of having a manual nbx_user.py file is now deprecated\n'
#       f'  Fix: nbx {method} upload file_path:fn_cls_name --id "id"'
#     )
#     raise ValueError("Old style upload is not supported anymore")

#   # In order to upload we can either chose to upload the entire folder, but can we implement a style where
#   # only that specific function is uploaded? This is useful for raw distributed compute style.
#   commands = init_folder.split(":")
#   if len(commands) == 2:
#     fn_file, fn_name = commands
#     mode = "folder"
#   elif len(commands) == 3:
#     mode, fn_file, fn_name = commands
#     if mode not in ["file", "folder"]:
#       raise ValueError(f"Invalid mode: '{mode}' in upload command, should be either 'file' or 'folder'")
#   else:
#     raise ValueError(f"Invalid init_folder: {init_folder}")
#   if mode != "folder":
#     raise NotImplementedError(f"Only folder mode is supported, got: {mode}")
#   if not os.path.exists(fn_file+".py"):
#     raise ValueError(f"File {fn_file}.py does not exist")
#   init_folder, file_name = os.path.split(fn_file)
#   init_folder = init_folder or "."
#   fn_name = fn_name.strip()
#   if not os.path.exists(init_folder):
#     raise ValueError(f"Folder {init_folder} does not exist")
#   logger.info(f"Uploading code from folder: {init_folder}:{file_name}:{fn_name}")
#   _curdir = os.getcwd()
#   os.chdir(init_folder)

#   workspace_id = workspace_id or secret.workspace_id

#   # Now that we have the init_folder and function name, we can throw relevant errors
#   perform_tea = True
#   if method == OT.SERVING:
#     if serving_type not in SST.all():
#       raise ValueError(f"Invalid serving_type: {serving_type}, should be one of {SST.all()}")
#     if serving_type == SST.FASTAPI or serving_type == SST.FASTAPI_V2:
#       logger.warning(f"You have selected serving_type='{serving_type}', this assumes the object: {fn_name} is a FastAPI app")
#       init_code = fn_name
#       perform_tea = False
#       load_operator = False

#   if perform_tea:
#     # build an Astea and analyse it for getting the computation that is going to be run
#     load_operator = True
#     tea = Astea(fn_file+".py")
#     items = tea.find(fn_name, [IT.CLASS, IT.FUNCTION])
#     if len(items) > 1:
#       raise ModuleNotFoundError(f"Multiple {fn_name} found in {fn_file}.py")
#     elif len(items) == 0:
#       logger.error(f"Could not find function or class type: '{fn_name}'")
#       raise ModuleNotFoundError(f"Could not find function or class type: '{fn_name}'")
#     fn = items[0]
#     if fn.type == IT.FUNCTION:
#       # does not require initialisation
#       if len(init_kwargs):
#         logger.error(
#           f"Cannot pass kwargs to a function: '{fn_name}'\n"
#           f"  Fix: you cannot pass kwargs {set(init_kwargs.keys())} to a function"
#         )
#         raise ValueError("Function does not require initialisation")
#       init_code = fn_name
#     elif fn.type == IT.CLASS:
#       # requires initialisation, in this case we will store the relevant to things in a Relic
#       init_comm = ",".join([f"{k}={v}" for k, v in init_kwargs.items()])
#       init_code = f"{fn_name}({init_comm})"
#       logger.info(f"Starting with init code:\n  {init_code}")

#   # load up the things that are to be passed to the exe.py file
#   exe_jinja_kwargs = {
#     "file_name": file_name,
#     "fn_name": fn_name,
#     "init_code": init_code,
#     "load_operator": load_operator,
#     "serving_type": serving_type,
#   }

#   # create a requirements.txt file if it doesn't exist with the latest nbox version
#   if not os.path.exists(U.join(".", "requirements.txt")):
#     with open(U.join(".", "requirements.txt"), "w") as f:
#       f.write(f'nbox[serving]=={__version__}\n')

#   # create a .nboxignore file if it doesn't exist, this will tell all the files not to be uploaded to the job pod
#   _igp = set(ospec.FN_IGNORE)
#   if not os.path.exists(U.join(".", ".nboxignore")):
#     with open(U.join(".", ".nboxignore"), "w") as f:
#       f.write("\n".join(_igp))
#   else:
#     with open(U.join(".", ".nboxignore"), "r") as f:
#       _igp = _igp.union(set(f.read().splitlines()))
#     _igp = sorted(list(_igp)) # sort it so it doesn't keep creating diffs in the git
#     with open(U.join(".", ".nboxignore"), "w") as f:
#       f.write("\n".join(_igp))

#   # creation of resources, we first need to check if any resource arguments are passed, if they are
#   def __common_resource(db: Resource) -> Resource:
#     # get a common resource based on what the user has said, what the db has and defaults if nothing is given
#     logger.debug(lo("db resources", **mpb.message_to_dict(db)))
#     resource = Resource(
#       cpu = str(resource_cpu) or db.cpu or ospec.DEFAULT_RESOURCE.cpu,
#       memory = str(resource_memory) or db.memory or ospec.DEFAULT_RESOURCE.memory,
#       disk_size = str(resource_disk_size) or db.disk_size or ospec.DEFAULT_RESOURCE.disk_size,
#       gpu = str(resource_gpu) or db.gpu or ospec.DEFAULT_RESOURCE.gpu,
#       gpu_count = str(resource_gpu_count) or db.gpu_count or ospec.DEFAULT_RESOURCE.gpu_count,
#       timeout = int(resource_timeout) or db.timeout or ospec.DEFAULT_RESOURCE.timeout,
#       max_retries = int(resource_max_retries) or db.max_retries or ospec.DEFAULT_RESOURCE.max_retries,
#     )
#     return resource

#   # common to both, kept out here because these two will eventually merge
#   nbx_auth_info = auth_info_pb()
#   if method == ospec.OperatorType.JOB:
#     # since user has not passed any arguments, we will need to check if the job already exists
#     job_proto: JobProto = nbox_grpc_stub.GetJob(
#       JobRequest(
#         auth_info = nbx_auth_info,
#         job = JobProto(id = id)
#       )
#     )
#     resource = __common_resource(job_proto.resource)
#     logger.debug(lo("resources", **mpb.message_to_dict(resource)))
#     out: Job = deploy_job(
#       init_folder = init_folder,
#       job_id = job_proto.id,
#       job_name = job_proto.name,
#       feature_gates = feature_gates,
#       dag = DAGProto(),
#       workspace_id = workspace_id,
#       schedule = None,
#       resource = resource,
#       exe_jinja_kwargs = exe_jinja_kwargs,
#     )
#     if trigger:
#       logger.info(f"Triggering job: {job_proto.name} ({job_proto.id})")
#       out = out.trigger()

#   elif method == ospec.OperatorType.SERVING:
#     model_name = model_name or U.get_random_name().replace("-", "_")
#     logger.info(f"Model name: {model_name}")

#     # serving_id, serving_name = _get_deployment_data(name = name, id = id, workspace_id = workspace_id)
#     serving_proto: Serving = nbox_serving_service_stub.GetServing(
#       ServingRequest(
#         auth_info = nbx_auth_info,
#         serving = Serving(id=id),
#       )
#     )
#     resource = __common_resource(serving_proto.resource)
#     logger.debug(lo("resources", **mpb.message_to_dict(resource)))
#     out: Serve = deploy_serving(
#       init_folder = init_folder,
#       serving_id = serving_proto.id,
#       model_name = model_name,
#       serving_name = serving_proto.name,
#       workspace_id = workspace_id,
#       resource = resource,
#       wait_for_deployment = False,
#       model_metadata = {
#         "serving_type": serving_type
#       },
#       exe_jinja_kwargs = exe_jinja_kwargs,
#     )
#     if deploy:
#       out.deploy(feature_gates = feature_gates, resource = resource)
#     if trigger:
#       out.pin()
#   else:
#     raise ValueError(f"Unknown method: {method}")

#   os.chdir(_curdir)

#   if _ret:
#     return out


# ################################################################################
# """
# # NimbleBox.ai Serving

# This is the proposed interface for the NimbleBox.ai Serving API. We want to keep
# the highest levels of consistency with the NBX-Jobs API.
# """
# ################################################################################

# @lru_cache()
# def _get_deployment_data(name: str = "", id: str = "", *, workspace_id: str = "") -> Tuple[str, str]:
#   """
#   Get the deployment data, either by name or id

#   Args:
#     name (str, optional): Name of the deployment. Defaults to "".
#     id (str, optional): ID of the deployment. Defaults to "".

#   Returns:
#     Tuple[str, str]: (id, name)
#   """
#   # print("Getting deployment data", name, id, workspace_id)
#   if (not name and not id) or (name and id):
#     logger.warning("Must provide either name or id")
#     return None, None
#   # filter and get "id" and "name"
#   workspace_id = workspace_id or secret.workspace_id

#   # get the deployment
#   serving: Serving = mpb.rpc(
#     nbox_serving_service_stub.GetServing,
#     ServingRequest(
#       serving=Serving(name=name, id=id),
#       auth_info = auth_info_pb()
#     ),
#     "Could not get deployment",
#     raise_on_error=True
#   )

#   return serving.id, serving.name

# def print_serving_list(sort: str = "created_on", *, workspace_id: str = ""):
#   """
#   Print the list of deployments

#   Args:
#     sort (str, optional): Sort by. Defaults to "created_on".
#   """
#   def _get_time(t):
#     return datetime.fromtimestamp(int(float(t))).strftime("%Y-%m-%d %H:%M:%S")

#   workspace_id = workspace_id or secret.workspace_id
#   all_deployments: ServingListResponse = mpb.rpc(
#     nbox_serving_service_stub.ListServings,
#     ServingListRequest(
#       auth_info=auth_info_pb(),
#       limit=10
#     ),
#     "Could not get deployments",
#     raise_on_error=True
#   )
#   # sorted_depls = sorted(all_deployments, key = lambda x: x[sort], reverse = sort == "created_on")
#   # headers = ["created_on", "id", "name", "pinned_id", "pinned_name", "pinned_last_updated"]
#   # [TODO] add sort by create time
#   headers = ["id", "name", "pinned_id", "pinned_name", "pinned_last_updated"]
#   all_depls = []
#   for depl in all_deployments.servings:
#     _depl = (depl.id, depl.name)
#     pinned = depl.models[0] if len(depl.models) else None
#     if not pinned:
#       _depl += (None, None,)
#     else:
#       _depl += (pinned.id, pinned.name, _get_time(pinned.created_at.seconds))
#     all_depls.append(_depl)

#   for l in tabulate.tabulate(all_depls, headers).splitlines():
#     logger.info(l)


# class Serve:
#   status = staticmethod(print_serving_list)
#   upload: 'Serve' = staticmethod(partial(upload_job_folder, "serving"))

#   def __init__(self, serving_id: str = "", model_id: str = "", *, workspace_id: str = "") -> None:
#     """Python wrapper for NBX-Serving gRPC API

#     Args:
#       serving_id (str, optional): Serving ID. Defaults to None.
#       model_id (str, optional): Model ID. Defaults to None.
#     """
#     self.id = serving_id
#     self.model_id = model_id
#     self.workspace_id = workspace_id or secret.workspace_id
#     if workspace_id is None:
#       raise DeprecationWarning("Personal workspace does not support serving")
#     else:
#       serving_id, serving_name = _get_deployment_data(name = "", id = self.id, workspace_id = self.workspace_id) # TODO add name support
#     self.serving_id = serving_id
#     self.serving_name = serving_name
#     self.ws_stub = nbox_ws_v1.deployments

#   def __repr__(self) -> str:
#     x = f"nbox.Serve('{self.id}', '{self.workspace_id}'"
#     if self.model_id is not None:
#       x += f", model_id = '{self.model_id}'"
#     x += ")"
#     return x

#   # there are things that are on the serving group level
  
#   def logs(self, f = sys.stdout):
#     """Get the logs of the model deployment

#     Args:
#       f (file, optional): File to write the logs to. Defaults to sys.stdout.
#     """
#     logger.debug(f"Streaming logs of model '{self.model_id}'")
#     for model_log in mpb.streaming_rpc(
#       nbox_model_service_stub.ModelLogs,
#       ModelRequest(
#         model = ModelProto(
#           id = self.model_id,
#           serving_group_id = self.serving_id
#         ),
#         auth_info = auth_info_pb(),
#       ),
#       f"Could not get logs of model {self.model_id}, only live logs are available",
#       False
#     ):
#       for log in model_log.log:
#         f.write(log)
#         f.flush()

#   def change_serving_resources(self, resource: Dict[str, Any]):
#     if type(resource, Resource):
#       pass
#     elif type(resource, dict):
#       resource = mpb.dict_to_message(resource, Resource())
#     else:
#       raise TypeError(f"resource must be of type dict or Resource, not {type(resource)}")

#     logger.info(f"Updating resources of a serving group: {self.serving_id}")
#     mpb.rpc(
#       nbox_serving_service_stub.UpdateServing,
#       UpdateServingRequest(
#         model=Serving(
#           id=self.model_id,
#           resource=resource
#         ),
#         update_mask=FieldMask(paths=["resource"]),
#         auth_info = auth_info_pb()
#       ),
#       "Could not change resources of serving group",
#       raise_on_error = True
#     )
#     pass

#   # there are things on the model level

#   def deploy(self, tag: str = "", feature_gates: Dict[str, str] = {}, resource: Dict[str, Any] = {}):
#     if not self.model_id:
#       raise ValueError("Model ID is required")

#     if type(resource) == Resource:
#       pass
#     elif type(resource) == dict:
#       resource = mpb.dict_to_message(resource, Resource())
#     else:
#       raise TypeError(f"resource must be of type dict or Resource, not {type(resource)}")

#     model = ModelProto(
#       id = self.model_id,
#       serving_group_id = self.serving_id,
#       resource = resource
#     )
#     if tag:
#       model.feature_gates.update({"SetModelMetadata": tag})
#     if feature_gates:
#       model.feature_gates.update(feature_gates)
#     logger.info(f"Deploying model {self.model_id} to deployment {self.serving_id} with tag: '{tag}' and feature gates: {feature_gates}")
#     try:
#       nbox_model_service_stub.Deploy(ModelRequest(model = model, auth_info = auth_info_pb()))
#     except grpc.RpcError as e:
#       logger.error(lo(
#         f"Could not deploy model {self.model_id} to deployment {self.serving_id}\n",
#         f"gRPC Code: {e.code()}\n"
#         f"    Error: {e.details()}",
#       ))

#   def pin(self) -> bool:
#     """Pin a model to the deployment

#     Args:
#       model_id (str, optional): Model ID. Defaults to None.
#       workspace_id (str, optional): Workspace ID. Defaults to "".
#     """
#     if not self.model_id:
#       raise ValueError("Model ID is required")

#     logger.info(f"Pin model {self.model_id} to deployment {self.serving_id}")
#     mpb.rpc(
#       nbox_model_service_stub.SetModelPin,
#       ModelRequest(
#         model = ModelProto(
#           id = self.model_id,
#           serving_group_id = self.serving_id,
#           pin_status = ModelProto.PinStatus.PIN_STATUS_PINNED
#         ),
#         auth_info = auth_info_pb()
#       ),
#       "Could not pin model",
#       raise_on_error = True
#     )
  
#   def unpin(self) -> bool:
#     """Pin a model to the deployment

#     Args:
#       model_id (str, optional): Model ID. Defaults to None.
#       workspace_id (str, optional): Workspace ID. Defaults to "".
#     """
#     if not self.model_id:
#       raise ValueError("Model ID is required")

#     logger.info(f"Unpin model {self.model_id} to deployment {self.serving_id}")
#     mpb.rpc(
#       nbox_model_service_stub.SetModelPin,
#       ModelRequest(
#         model = ModelProto(
#           id = self.model_id,
#           serving_group_id = self.serving_id,
#           pin_status = ModelProto.PinStatus.PIN_STATUS_UNPINNED
#         ),
#         auth_info = auth_info_pb(),
#       ),
#       "Could not unpin model",
#       raise_on_error = True
#     )
  
#   def scale(self, replicas: int) -> bool:
#     """Scale the model deployment

#     Args:
#       replicas (int): Number of replicas
#     """
#     if not self.model_id:
#       raise ValueError("Model ID is required to scale a model. Pass with --model_id")
#     if replicas < 0:
#       raise ValueError("Replicas must be greater than or equal to 0")

#     logger.info(f"Scale model deployment {self.model_id} to {replicas} replicas")
#     mpb.rpc(
#       nbox_model_service_stub.UpdateModel,
#       UpdateModelRequest(
#         model=ModelProto(
#           id=self.model_id,
#           serving_group_id=self.serving_id,
#           replicas=replicas
#         ),
#         update_mask=FieldMask(paths=["replicas"]),
#         auth_info = auth_info_pb()
#       ),
#       "Could not scale deployment",
#       raise_on_error = True
#     )

#   def change_model_resources(self, resource: Dict[str, Any]):
#     if type(resource, Resource):
#       pass
#     elif type(resource, dict):
#       resource = mpb.dict_to_message(resource, Resource())
#     else:
#       raise TypeError(f"resource must be of type dict or Resource, not {type(resource)}")

#     logger.warn(f"Updating resources of a running model '{self.model_id}', this will result in a restart!")
#     mpb.rpc(
#       nbox_model_service_stub.UpdateModel,
#       UpdateModelRequest(
#         model=ModelProto(
#           id=self.model_id,
#           serving_group_id=self.serving_id,
#           resource=resource
#         ),
#         update_mask=FieldMask(paths=["resource"]),
#         auth_info = auth_info_pb()
#       ),
#       "Could not change resources of a running model",
#       raise_on_error = True
#     )


# ################################################################################
# """
# # NimbleBox.ai Jobs

# This is the actual job object that users can manipulate. It is a shallow class
# around the NBX-Jobs gRPC API.
# """
# ################################################################################

# @lru_cache()
# def _get_job_data(name: str = "", id: str = "", remove_archived: bool = True, *, workspace_id: str = "") -> Tuple[str, str]:
#   """
#   Returns the job id and name
  
#   Args:
#     name (str, optional): Job name. Defaults to "".
#     id (str, optional): Job ID. Defaults to "".
#     remove_archived (bool, optional): If True, will remove archived jobs. Defaults to True.

#   Returns:
#     Tuple[str, str]: Job ID, Job Name
#   """
#   if (not name and not id) or (name and id):
#     logger.info(f"Please either pass job_id '{id}' or name '{name}'")
#     return None, None
#   # get stub
#   workspace_id = workspace_id or secret.workspace_id
#   if workspace_id == None:
#     workspace_id = "personal"

#   job: JobProto = mpb.rpc(
#     nbox_grpc_stub.GetJob,
#     JobRequest(
#       auth_info = auth_info_pb(),
#       job = JobProto(id=id, name=name)
#     ),
#     "Could not find job with ID: {}".format(id),
#     raise_on_error = True
#   )
#   if job.status == JobProto.Status.ARCHIVED and remove_archived:
#     logger.info(f"Job {job.id} is archived. Will try to create a new Job.")
#     return None, name
#   job_name = job.name
#   job_id = job.id
#   logger.info(f"Found job with ID '{job_id}' and name '{job_name}'")
#   return job_id, job_name

# def get_job_list(sort: str = "name", *, workspace_id: str = ""):
#   """Get list of jobs
  
#   Args:
#     sort (str, optional): Sort key. Defaults to "name".
#   """
#   workspace_id = workspace_id or secret.workspace_id

#   def _get_time(t):
#     return datetime.fromtimestamp(int(float(t))).strftime("%Y-%m-%d %H:%M:%S")

#   out: ListJobsResponse = mpb.rpc(
#     nbox_grpc_stub.ListJobs,
#     ListJobsRequest(auth_info = auth_info_pb()),
#     "Could not get job list",
#   )

#   if len(out.jobs) == 0:
#     logger.info("No jobs found")
#     sys.exit(0)

#   headers = ['created_at', 'id', 'name', 'schedule', 'status']
#   try:
#     sorted_jobs = sorted(out.jobs, key = lambda x: getattr(x, sort))
#   except Exception as e:
#     logger.error(f"Cannot sort on key: {sort}")
#     sorted_jobs = out.jobs
#   data = []
#   for j in sorted_jobs:
#     _row = []
#     for x in headers:
#       if x == "status":
#         _row.append(JobProto.Status.keys()[getattr(j, x)])
#         continue
#       elif x == "created_at":
#         _row.append(_get_time(j.created_at.seconds))
#       elif x == "schedule":
#         _row.append(j.schedule.cron)
#       else:
#         _row.append(getattr(j, x))
#     data.append(_row)
#   for l in tabulate.tabulate(data, headers).splitlines():
#     logger.info(l)


# class Job:
#   status = staticmethod(get_job_list)
#   upload: 'Job' = staticmethod(partial(upload_job_folder, "job"))

#   def __init__(self, job_name: str = "", job_id: str = ""):
#     """Python wrapper for NBX-Jobs gRPC API, when both arguments are not passed,
#     an unintialiased object is created.

#     Args:
#       job_name (str, optional): Job name. Defaults to "".
#       job_id (str, optional): Job ID. Defaults to "".
#     """
#     if job_name == "" and job_id == "":
#       return

#     self.id, self.name = _get_job_data(job_name, job_id)
#     self.workspace_id = secret.workspace_id
#     self.auth_info = auth_info_pb()
#     self.job_proto = JobProto(id = self.id)

#     self.run_stub = None
#     self.runs = []

#     # sometimes a Job can be a placeholder and not have any data against it, thus it won't make
#     # sense to try to get the data. This causes RPC error
#     if self.id is not None:
#       self.refresh()
#       self.get_runs() # load the runs as well

#   @property
#   def exists(self):
#     """Check if this job exists in the workspace"""
#     return self.id is not None

#   @classmethod
#   def new(cls, job_name: str, description: str = ""):
#     """Create a new job

#     Args:
#       job_name (str): Job name
#       description (str, optional): Job description. Defaults to "".

#     Returns:
#       Job: Job object
#     """
#     job = nbox_ws_v1.job("post", job_name = job_name, job_description = description)
#     job_id = job["job_id"]
#     # job_proto: JobProto = nbox_grpc_stub.CreateJob(
#     #   JobRequest(
#     #     auth_info = auth_info_pb(),
#     #     job = JobProto(
#     #       name = job_name,
#     #       code = Code(
#     #         size = 1,
#     #         type = Code.Type.NOT_SET,
#     #       ),
#     #       resource = DEFAULT_RESOURCE,
#     #       schedule = JobProto.Schedule(
#     #         cron = "0 0 * * *",
#     #         end = Timestamp(seconds = 0),
#     #       ),
#     #       status = JobProto.Status.NOT_SET,
#     #       paused = False,
#     #     ),
#     #   )
#     # )
#     logger.info(f"Created a new job with name (ID): {job_name} ({job_id})")
#     return cls(job_id = job_id)

#   def change_schedule(self, new_schedule: Schedule):
#     """Change schedule this job
    
#     Args:
#       new_schedule (Schedule): New schedule
#     """
#     logger.debug(f"Updating job '{self.job_proto.id}'")
#     self.job_proto.schedule.MergeFrom(new_schedule.get_message())
#     mpb.rpc(
#       nbox_grpc_stub.UpdateJob,
#       UpdateJobRequest(auth_info=self.auth_info, job=self.job_proto, update_mask=FieldMask(paths=["schedule"])),
#       "Could not update job schedule",
#       raise_on_error = True
#     )
#     logger.debug(f"Updated job '{self.job_proto.id}'")
#     self.refresh()

#   def __repr__(self) -> str:
#     x = f"nbox.Job('{self.job_proto.id}', '{self.job_proto.auth_info.workspace_id}'): {self.status}"
#     if self.job_proto.schedule.ByteSize != None:
#       x += f" {self.job_proto.schedule}"
#     else:
#       x += " (no schedule)"
#     return x

#   def logs(self, f = sys.stdout):
#     """Stream logs of the job, `f` can be anything has a `.write/.flush` methods"""
#     logger.debug(f"Streaming logs of job '{self.job_proto.id}'")
#     for job_log in mpb.streaming_rpc(
#       nbox_grpc_stub.GetJobLogs,
#       JobRequest(auth_info=self.auth_info ,job = self.job_proto),
#       f"Could not get logs of job {self.job_proto.id}, is your job complete?",
#       True
#     ):
#       for log in job_log.log:
#         f.write(log)
#         f.flush()

#   def delete(self):
#     """Delete this job"""
#     logger.info(f"Deleting job '{self.job_proto.id}'")
#     mpb.rpc(nbox_grpc_stub.DeleteJob, JobRequest(auth_info=self.auth_info, job = self.job_proto,), "Could not delete job")
#     logger.info(f"Deleted job '{self.job_proto.id}'")
#     self.refresh()

#   def refresh(self):
#     """Refresh Job data"""
#     logger.info(f"Updating job '{self.job_proto.id}'")
#     if self.id == None:
#       self.id, self.name = _get_job_data(id = self.id, workspace_id = self.workspace_id)
#     if self.id == None:
#       return

#     self.job_proto: JobProto = mpb.rpc(
#       nbox_grpc_stub.GetJob,
#       JobRequest(auth_info=self.auth_info, job = self.job_proto),
#       f"Could not get job {self.job_proto.id}"
#     )
#     self.auth_info.CopyFrom(auth_info_pb())
#     logger.debug(f"Updated job '{self.job_proto.id}'")

#     self.status = self.job_proto.Status.keys()[self.job_proto.status]

#   def trigger(self, tag: str = "", feature_gates: dict = {}):
#     """Manually triger this job.
    
#     Args:
#       tag (str, optional): Tag to be set in the run metadata, read in more detail before trying to use this. Defaults to "".
#     """
#     logger.debug(f"Triggering job '{self.job_proto.id}'")
#     if tag:
#       self.job_proto.feature_gates.update(feature_gates)
#       self.job_proto.feature_gates.update({"SetRunMetadata": tag})
#     mpb.rpc(nbox_grpc_stub.TriggerJob, JobRequest(auth_info=self.auth_info, job = self.job_proto), f"Could not trigger job '{self.job_proto.id}'")
#     logger.info(f"Triggered job '{self.job_proto.id}'")
#     self.refresh()

#   def pause(self):
#     """Pause the execution of this job.

#     **WARNING: This will "cancel" all the scheduled runs, if present**
#     """
#     logger.info(f"Pausing job '{self.job_proto.id}'")
#     job: JobProto = self.job_proto
#     job.status = JobProto.Status.PAUSED
#     mpb.rpc(nbox_grpc_stub.UpdateJob, UpdateJobRequest(auth_info=self.auth_info, job=job, update_mask=FieldMask(paths=["status", "paused"])), f"Could not pause job {self.job_proto.id}", True)
#     logger.debug(f"Paused job '{self.job_proto.id}'")
#     self.refresh()

#   def resume(self):
#     """Resume the Job with the current schedule, if provided else simlpy sets status as ACTIVE"""
#     logger.info(f"Resuming job '{self.job_proto.id}'")
#     job: JobProto = self.job_proto
#     job.status = JobProto.Status.SCHEDULED
#     mpb.rpc(nbox_grpc_stub.UpdateJob, UpdateJobRequest(auth_info=self.auth_info, job=job, update_mask=FieldMask(paths=["status", "paused"])), f"Could not resume job {self.job_proto.id}", True)
#     logger.debug(f"Resumed job '{self.job_proto.id}'")
#     self.refresh()

#   def _get_runs(self, page = -1, limit = 10) -> List[Dict]:
#     self.run_stub = nbox_ws_v1.job.u(self.id).runs
#     runs = self.run_stub(limit = limit, page = page)["runs_list"]
#     return runs

#   def get_runs(self, page = -1, sort = "s_no", limit = 10) -> List[Dict]:
#     """
#     Get runs for this job

#     Args:
#       page (int, optional): Page number. Defaults to -1.
#       sort (str, optional): Sort by. Defaults to "s_no".
#       limit (int, optional): Number of runs to return. Defaults to 10.

#     Returns:
#       List[Dict]: List of runs
#     """
#     runs = self._get_runs(page, limit)
#     sorted_runs = sorted(runs, key = lambda x: x[sort])
#     return sorted_runs

#   def display_runs(self, sort: str = "created_at", page: int = -1, limit = 10):
#     """Display runs for this job

#     Args:
#       sort (str, optional): Sort by. Defaults to "created_at".
#       page (int, optional): Page number. Defaults to -1.
#       limit (int, optional): Number of runs to return. Defaults to 10.
#     """
#     headers = ["s_no", "created_at", "run_id", "status"]

#     def _display_runs(runs):
#       data = []
#       for run in runs:
#         data.append((run["s_no"], run["created_at"], run["run_id"], run["status"]))
#       for l in tabulate.tabulate(data, headers).splitlines():
#         logger.info(l)

#     runs = self.get_runs(sort = sort, page = page, limit = limit) # time should be reverse
#     _display_runs(runs)
#     if page == -1:
#       page = 1
#     y = input(f">> Print {limit} more runs? (y/n): ")
#     done = y != "y"
#     while not done:
#       _display_runs(self.get_runs(page = page + 1, sort = sort, limit = limit))
#       page += 1
#       y = input(f">> Print {limit} more runs? (y/n): ")
#       done = y != "y"

#   def last_n_runs(self, n: int = 10) -> List[Dict]:
#     """Get last n runs for this job

#     Args:
#       n (int, optional): Number of runs to return. Defaults to 10.

#     Returns:
#       List[Dict]: List of runs
#     """
#     all_items = []
#     _page = 1
#     out = self._get_runs(_page, n)
#     all_items.extend(out)

#     while len(all_items) < n:
#       _page += 1
#       out = self._get_runs(_page, n)
#       if not len(out):
#         break
#       all_items.extend(out)

#     if n == 1:
#       return all_items[0]
#     return all_items
